"""Logging pipeline runs, node output and status information in mara database"""

import psycopg2.extensions
import sqlalchemy.orm
from sqlalchemy.ext.declarative import declarative_base

import mara_db.postgresql
from .. import config
from ..logging import pipeline_events, system_statistics
from .. import events

Base = declarative_base()


class Run(Base):
    """Runtime and status of a data integration run"""
    __tablename__ = 'data_integration_run'

    run_id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    node_path = sqlalchemy.Column(sqlalchemy.ARRAY(sqlalchemy.TEXT), nullable=False, index=True)
    pid = sqlalchemy.Column(sqlalchemy.Integer, nullable=False)
    start_time = sqlalchemy.Column(sqlalchemy.TIMESTAMP(timezone=True), nullable=False)
    end_time = sqlalchemy.Column(sqlalchemy.TIMESTAMP(timezone=True))
    succeeded = sqlalchemy.Column(sqlalchemy.BOOLEAN)


class NodeRun(Base):
    """Runtime, status information and output of a pipeline node run"""
    __tablename__ = 'data_integration_node_run'

    node_run_id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    run_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('data_integration_run.run_id'), index=True)

    node_path = sqlalchemy.Column(sqlalchemy.ARRAY(sqlalchemy.TEXT), index=True)
    start_time = sqlalchemy.Column(sqlalchemy.TIMESTAMP(timezone=True), nullable=False)
    end_time = sqlalchemy.Column(sqlalchemy.TIMESTAMP(timezone=True))
    succeeded = sqlalchemy.Column(sqlalchemy.BOOLEAN)
    is_pipeline = sqlalchemy.Column(sqlalchemy.BOOLEAN)


class NodeOutput(Base):
    """Runtime, status information and output of a pipeline node run"""
    __tablename__ = 'data_integration_node_output'

    node_output_id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    node_run_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('data_integration_node_run.node_run_id'),
                                    index=True)
    timestamp = sqlalchemy.Column(sqlalchemy.TIMESTAMP(timezone=True), nullable=False)
    message = sqlalchemy.Column(sqlalchemy.TEXT, nullable=False)
    format = sqlalchemy.Column(sqlalchemy.TEXT, nullable=False)
    is_error = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False)


class SystemStatistics(Base):
    """System stats measurements"""
    __tablename__ = 'data_integration_system_statistics'

    timestamp = sqlalchemy.Column(sqlalchemy.TIMESTAMP(timezone=True), primary_key=True, index=True)
    disc_read = sqlalchemy.Column(sqlalchemy.FLOAT)
    disc_write = sqlalchemy.Column(sqlalchemy.FLOAT)
    net_recv = sqlalchemy.Column(sqlalchemy.FLOAT)
    net_sent = sqlalchemy.Column(sqlalchemy.FLOAT)
    cpu_usage = sqlalchemy.Column(sqlalchemy.FLOAT)
    mem_usage = sqlalchemy.Column(sqlalchemy.FLOAT)
    swap_usage = sqlalchemy.Column(sqlalchemy.FLOAT)
    iowait = sqlalchemy.Column(sqlalchemy.FLOAT)


def close_open_run_after_error(run_id: int):
    """Closes all open run and node_run for this run_id as failed"""
    if run_id is None:
        return
    print(f'Run aborted, cleaning up (run_id = {run_id})')
    _close_run = f'''
UPDATE  data_integration_run
SET end_time = now(), succeeded = FALSE
WHERE run_id = {"%s"} and end_time IS NULL
    '''
    _close_node_run = f'''
UPDATE  data_integration_node_run
SET end_time = now(), succeeded = FALSE
WHERE run_id = {"%s"} and end_time IS NULL
        '''
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
        cursor.execute(_close_node_run, (run_id,))
        cursor.execute(_close_run, (run_id,))


class RunLogger(events.EventHandler):
    run_id: int = None
    node_output: {tuple: [pipeline_events.Output]} = None

    def handle_event(self, event: events.Event):

        if isinstance(event, pipeline_events.RunStarted):
            with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
                cursor.execute(f'''
INSERT INTO data_integration_run (node_path, pid, start_time)
VALUES ({"%s, %s, %s"})
RETURNING run_id;''', (event.node_path, event.pid, event.start_time))
                self.run_id = cursor.fetchone()[0]

        elif isinstance(event, pipeline_events.Output):
            key = tuple(event.node_path)

            if not self.node_output:
                self.node_output = {}

            if key in self.node_output:
                self.node_output[key].append(event)
            else:
                self.node_output[key] = [event]

        elif isinstance(event, pipeline_events.NodeStarted):
            with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
                cursor.execute(f'''
INSERT INTO data_integration_node_run (run_id, node_path, start_time, is_pipeline)
VALUES  ({"%s, %s, %s, %s"})
RETURNING node_run_id''', (self.run_id, event.node_path, event.start_time, event.is_pipeline))

        elif isinstance(event, system_statistics.SystemStatistics):
            with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
                cursor.execute(f'''
INSERT INTO data_integration_system_statistics (timestamp, disc_read, disc_write, net_recv, net_sent,
                                  cpu_usage, mem_usage, swap_usage, iowait)
VALUES ({"%s, %s, %s, %s, %s, %s, %s, %s, %s"})''',
                               (event.timestamp, event.disc_read, event.disc_write, event.net_recv,
                                event.net_sent, event.cpu_usage, event.mem_usage, event.swap_usage, event.iowait))

        elif isinstance(event, pipeline_events.NodeFinished):
            with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
                cursor.execute(f'''
UPDATE data_integration_node_run
SET end_time={"%s"}, succeeded={"%s"}
WHERE run_id={"%s"} AND node_path={"%s"}
RETURNING node_run_id''', (event.end_time, event.succeeded, self.run_id, event.node_path))
                node_run_id = cursor.fetchone()[0]

                cursor.execute('''
INSERT INTO data_integration_node_output (node_run_id, timestamp, message, format, is_error)
VALUES ''' + ','.join([cursor.mogrify('(%s,%s,%s,%s,%s)', (node_run_id, output_event.timestamp, output_event.message,
                                                           output_event.format, output_event.is_error))
                      .decode('utf-8')
                       for output_event in self.node_output.get(tuple(event.node_path))]))

        elif isinstance(event, pipeline_events.RunFinished):
            with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
                cursor.execute(f'''
UPDATE data_integration_run
SET end_time={"%s"}, succeeded={"%s"}
WHERE run_id={"%s"}''', (event.end_time, event.succeeded, self.run_id))

                cursor.execute(f'''
DELETE FROM data_integration_node_output WHERE node_run_id IN (
    SELECT node_run_id FROM data_integration_node_run WHERE run_id IN (
        SELECT run_id FROM data_integration_run
        WHERE start_time + INTERVAL '{config.run_log_retention_in_days()} days' < current_timestamp));''')

                cursor.execute(f'''
DELETE FROM data_integration_node_run WHERE run_id IN (
    SELECT run_id FROM data_integration_run
    WHERE start_time + INTERVAL '{config.run_log_retention_in_days()} days' < current_timestamp);''')

                cursor.execute(f'''
DELETE FROM data_integration_run
WHERE start_time + INTERVAL '{config.run_log_retention_in_days()} days' < current_timestamp;''')

                cursor.execute(f'''
DELETE FROM data_integration_system_statistics
WHERE timestamp + INTERVAL '{config.run_log_retention_in_days()} days' < current_timestamp;''')



"""Tracks the last comparison value of an incremental copy"""

import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base

import mara_db.config
import mara_db.dbs
import mara_db.postgresql

Base = declarative_base()


class IncrementalCopyStatus(Base):
    """The last `modification_value` for a table that is incrementally loaded"""
    __tablename__ = 'data_integration_incremental_copy_status'

    node_path = sqlalchemy.Column(sqlalchemy.ARRAY(sqlalchemy.Text), primary_key=True, index=True)
    source_table = sqlalchemy.Column(sqlalchemy.Text, primary_key=True)
    last_comparison_value = sqlalchemy.Column(sqlalchemy.Text)


def update(node_path: [str], source_db_alias: str, source_table: str, last_comparison_value):
    """
    Updates the last_comparison_value for a pipeline node and table
    Args:
        node_path: The path of the parent pipeline node
        source_db_alias: The alias of the the db from which data is copied
        source_table: The table from which is copied
        max_modification_value: The last retrieved modification value

    Returns:

    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f'''
INSERT INTO data_integration_incremental_copy_status (node_path, source_table, last_comparison_value)
VALUES ({'%s,%s,%s'})
ON CONFLICT (node_path, source_table)
DO UPDATE SET last_comparison_value = EXCLUDED.last_comparison_value
''', (node_path, f'{source_db_alias}.{source_table}', last_comparison_value))


def delete(node_path: [str], source_db_alias: str, source_table: str):
    """
    Deletes the last_comparison_value for a pipeline node and table
    Args:
        node_path: The path of the parent pipeline node
        source_db_alias: The alias of the the db from which data is copied
        source_table: The table from which is copied
    Returns:

    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f'''
DELETE FROM data_integration_incremental_copy_status
WHERE node_path = {'%s'} AND source_table = {'%s'}
''', (node_path, f'{source_db_alias}.{source_table}'))



def get_last_comparison_value(node_path: [str], source_db_alias: str, source_table: str):
    """
    Returns the last comparison value for a pipeline node and table
    Args:
        node_path: The path of the parent pipeline node
        source_db_alias: The alias of the the db from which data is copied
        source_table: The table from which is copied

    Returns:
        The value or None
    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f"""
SELECT last_comparison_value
FROM data_integration_incremental_copy_status
WHERE node_path = {'%s'} AND source_table = {'%s'}""", (node_path, f'{source_db_alias}.{source_table}'))
        result = cursor.fetchone()
        return result[0] if result else None



"""Visualization of pipelines using graphviz"""

import functools

import flask

from mara_page import acl, html, bootstrap, _
from . import views
from .. import pipelines


def card(node: pipelines.Pipeline):
    return bootstrap.card(
        body=[_.p[_.em[node.description]],
              (_.p[views.format_labels(node)] if node.labels else ''),
              html.asynchronous_content(
                  flask.url_for('data_integration.dependency_graph', path=node.url_path()))])


@views.blueprint.route('/<path:path>/dependency-graph')
@views.blueprint.route('/dependency-graph', defaults={'path': ''})
@acl.require_permission(views.acl_resource, do_abort=False)
@functools.lru_cache(maxsize=None)
def dependency_graph(path: str):
    node, found = pipelines.find_node(path.split('/'))
    if not found:
        flask.abort(404, f'Node "{path}" not found')

    return dependency_graph(node)


@functools.singledispatch
def dependency_graph(nodes: {str: pipelines.Node},
                     current_node: pipelines.Node = None) -> str:
    """
    Draws a list of pipeline nodes and the dependencies between them using graphviz

    Args:
        nodes: The nodes to render
        current_node: If not null, then this node is highlighted

    Returns:
        An svg representation of the graph
    """
    import graphviz

    graph = graphviz.Digraph(graph_attr={'rankdir': 'TD', 'ranksep': '0.25', 'nodesep': '0.1'})

    # for finding redundant edges
    @functools.lru_cache(maxsize=None)
    def all_transitive_upstream_ids(node: pipelines.Node, first_level: bool = True):
        upstream_ids = set()
        for upstream in node.upstreams:
            if not first_level:
                upstream_ids.add(upstream.id)
            upstream_ids.update(all_transitive_upstream_ids(upstream, False))
        return upstream_ids

    for node in nodes.values():
        node_attributes = {'fontname': ' ',  # use website default
                           'fontsize': '10.5px'  # fontsize unfortunately must be set
                           }

        if node != current_node:
            node_attributes.update(
                {'href': views.node_url(node), 'fontcolor': '#0275d8',
                 'tooltip': node.description, 'color': 'transparent'})
        else:
            node_attributes.update({'color': '#888888', 'style': 'dotted'})

        if isinstance(node, pipelines.Pipeline):
            node_attributes.update({'shape': 'rectangle', 'style': 'dotted', 'color': '#888888'})
        elif isinstance(node, pipelines.ParallelTask):
            node_attributes.update({'shape': 'ellipse', 'style': 'dotted', 'color': '#888888'})
        else:
            node_attributes['shape'] = 'rectangle'

        graph.node(name=node.id, label=node.id.replace('_', '\n'), _attributes=node_attributes)

        for upstream in node.upstreams:
            if upstream.id in nodes:
                style = 'solid'
                color = '#888888'

                if upstream.id in all_transitive_upstream_ids(node):
                    style = 'dashed'
                    color = '#cccccc'
                graph.edge(upstream.id, node.id,
                           _attributes={'color': color, 'arrowsize': '0.7', 'style': style})
            elif (not current_node) or node in current_node.upstreams:
                graph.node(name=f'{upstream.id}_{node.id}',
                           _attributes={'style': 'invis', 'label': '', 'height': '0.1', 'fixedsize': 'true'})
                graph.edge(f'{upstream.id}_{node.id}', node.id,
                           _attributes={'color': '#888888', 'arrowsize': '0.7',
                                        'edgetooltip': upstream.id, 'style': 'dotted'})

        for downstream in node.downstreams:
            if downstream.id not in nodes and (not current_node or node in current_node.downstreams):
                graph.node(name=f'{downstream.id}_{node.id}',
                           _attributes={'style': 'invis', 'label': '', 'height': '0.1', 'fixedsize': 'true'})
                graph.edge(node.id, f'{downstream.id}_{node.id}',
                           _attributes={'color': '#888888', 'arrowsize': '0.7',
                                        'edgetooltip': downstream.id, 'style': 'dotted'})

    try:
        return graph.pipe('svg').decode('utf-8')
    except graphviz.backend.ExecutableNotFound as e:
        return str(_.tt(style='color:red')[str(e)])


@dependency_graph.register(pipelines.Pipeline)
def __(pipeline: pipelines.Pipeline):
    """Draw all nodes of a pipeline excluding initial and final node"""
    return dependency_graph({id: node for id, node in pipeline.nodes.items()
                             if node != pipeline.initial_node and node != pipeline.final_node})


@dependency_graph.register(pipelines.Node)
def __(node: pipelines.Pipeline):
    """For all other pipeline nodes, draw only direct upstreams and downstreams of node"""
    return dependency_graph({node.id: node for node in list(node.upstreams) + [node] + list(node.downstreams)},
                            node)



import inspect
import re
import typing
from html import escape

from mara_page import _, html
from .. import pipelines
from ..commands import python


class ParallelExecutePython(pipelines.ParallelTask):
    def __init__(self, id: str, description: str, file_name: str, parameter_function: typing.Callable,
                 max_number_of_parallel_tasks: int = None, commands_before: [pipelines.Command] = None,
                 commands_after: [pipelines.Command] = None) -> None:
        super().__init__(id=id, description=description, max_number_of_parallel_tasks=max_number_of_parallel_tasks,
                         commands_before=commands_before, commands_after=commands_after)
        self.file_name = file_name
        self.parameter_function = parameter_function

    def add_parallel_tasks(self, sub_pipeline: 'pipelines.Pipeline') -> None:
        parameters = self.parameter_function()

        if not isinstance(parameters, list) or not all(isinstance(item, tuple) for item in parameters):
            raise ValueError(f'parameter function should return a list of tuples, got "{repr(parameters)}"')

        for parameter_tuple in parameters:
            sub_pipeline.add(pipelines.Task(
                id='_'.join([re.sub('[^0-9a-z\-_]+', '', str(x).lower().replace('-', '_')) for x in parameter_tuple]),
                description=f'Runs the script with parameters {repr(parameter_tuple)}',
                commands=[python.ExecutePython(file_name=self.file_name, args=list(parameter_tuple))]))

    def html_doc_items(self) -> [(str, str)]:
        path = self.parent.base_path() / self.file_name
        return [('parameter function',
                 html.highlight_syntax(inspect.getsource(self.parameter_function), 'python')),
                ('file name', _.i[self.file_name]),
                (_.i['file content'], html.highlight_syntax(path.read_text().strip('\n')
                                                            if self.file_name and path.exists()
                                                            else '', 'python'))]


class ParallelRunFunction(pipelines.ParallelTask):
    def __init__(self, id: str, description: str, function: typing.Callable, parameter_function: typing.Callable,
                 max_number_of_parallel_tasks: int = None, commands_before: [pipelines.Command] = None,
                 commands_after: [pipelines.Command] = None) -> None:
        super().__init__(id=id, description=description, max_number_of_parallel_tasks=max_number_of_parallel_tasks,
                         commands_before=commands_before, commands_after=commands_after)
        self.function = function
        self.parameter_function = parameter_function

    def add_parallel_tasks(self, sub_pipeline: 'pipelines.Pipeline') -> None:
        parameters = self.parameter_function()

        if not isinstance(parameters, list):
            raise ValueError(f'parameter function should return a list, got "{repr(parameters)}"')

        for parameter in parameters:
            sub_pipeline.add(pipelines.Task(
                id=str(parameter).lower().replace(' ', '_').replace('-', '_'),
                description=f'Runs the function with parameters {repr(parameter)}',
                commands=[python.RunFunction(lambda args=parameter: self.function(args))]))

    def html_doc_items(self) -> [(str, str)]:
        return [('function', _.pre[escape(str(self.function))]),
                ('parameter function',
                 html.highlight_syntax(inspect.getsource(self.parameter_function), 'python')),
                (_.i['implementation'], html.highlight_syntax(inspect.getsource(self.function), 'python'))]



import json
import abc
import datetime


class Event():
    def __init__(self) -> None:
        """
        Base class for events that are emitted from mara.
        """
        pass

    def to_json(self):
        return json.dumps({field: value.isoformat() if isinstance(value, datetime.datetime) else value
                           for field, value in self.__dict__.items()})


class EventHandler(abc.ABC):
    @abc.abstractmethod
    def handle_event(self, event: Event):
        pass


def notify_configured_event_handlers(event: Event):
    from . import config
    all_handlers = config.event_handlers()
    for handler in all_handlers:
        handler.handle_event(event)



"""Text output logging with redirection to queues"""

import multiprocessing
import sys
from datetime import datetime

from ..logging import pipeline_events
import data_integration.config

Format = pipeline_events.Output.Format


def log(message: str, format: pipeline_events.Output.Format = Format.STANDARD,
        is_error: bool = False) -> None:
    """
    Logs text messages.

    When run inside a pipeline, this will send a log message to the parent process.
    Otherwise, messages will be printed to `sys.stdout` and `sys.stderr`.

    Any string in `data_integration.config.password_masks()` will be replaced by '***'.

    Args:
        message: The message to display
        format: How to format the message
        is_error: Whether the message is considered an error message
    """
    message = message.rstrip()
    masks = data_integration.config.password_masks()
    if masks:
        for mask in masks:
            message = message.replace(mask, '***')
    if message:
        if _event_queue:
            _event_queue.put(pipeline_events.Output(_node_path, message, format, is_error))
        elif is_error:
            sys.stderr.write(message + '\n')
        else:
            sys.stdout.write(message + '\n')


_event_queue: multiprocessing.Queue = None
"""When running in a forked process, this will be bound to a queue for sending events to the parent process."""

_current_node_path = None
"""When running in a forked task process, this will be bound to the path of the currently running pipeline node"""


def redirect_output(event_queue: multiprocessing.Queue, node_path: [str]):
    """
    Redirects the output of the `log` function as well as `sys.stdout` and `sys.stderr` to `event_queue`
    Args:
        queue: The queue where to redirect messages to
        node_path: The id and parent ids of the currently running tasks
    """
    global _event_queue
    _event_queue = event_queue

    global _node_path
    _node_path = node_path

    class OutputRedirector():
        def __init__(self, is_error: bool) -> None:
            self.is_error = is_error

        def write(self, message):
            log(message=message, format=Format.VERBATIM, is_error=self.is_error)

        def flush(self):
            pass

    sys.stdout = OutputRedirector(is_error=False)
    sys.stderr = OutputRedirector(is_error=True)


def format_time_difference(t1: datetime, t2: datetime):
    """
    Displays the time difference from t1 to t2 in a human - readable form.
    Inspired by https://stackoverflow.com/a/11157649/243519
    """
    import dateutil.relativedelta

    difference = dateutil.relativedelta.relativedelta(t2, t1)
    return ', '.join([str(getattr(difference, attr)) + ' ' + attr for attr in
                      ['years', 'months', 'days', 'hours', 'minutes', 'seconds']
                      if getattr(difference, attr) or attr == 'seconds'])


if __name__ == "__main__":
    # normal case when not run from a pipeline execution
    log(message='hello world')  # -> stdout
    log(message='oops', is_error=True)  # -> stderr



import json
import pathlib

import flask
import psycopg2.extensions

import mara_db.config
import mara_db.postgresql
from mara_page import acl, bootstrap, html, _
from . import views
from .. import pipelines


def card(node: pipelines.Node):
    """A card that shows the duration of the node and its top children over time"""
    return bootstrap.card(
        header_left='Run times',
        body=html.asynchronous_content(
            flask.url_for('data_integration.run_time_chart', path=node.url_path())))


@views.blueprint.route('/<path:path>/run-time-chart')
@views.blueprint.route('/run-time-chart', defaults={'path': ''})
@acl.require_permission(views.acl_resource, do_abort=False)
def run_time_chart(path: str):
    node, found = pipelines.find_node(path.split('/'))
    if not found:
        flask.abort(404, f'Node "{path}" not found')

    query = (pathlib.Path(__file__).parent / 'run_time_chart.sql').read_text()

    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
        cursor.execute(query)
        cursor.execute(f'SELECT row_to_json(t) FROM pg_temp.node_run_times({"%s"}) t', (node.path(),))
        rows = [row[0] for row in cursor.fetchall()]

        if rows and len(rows) > 1:
            number_of_child_runs = len(rows[0]['child_runs']) if rows[0]['child_runs'] else 0

            return str(_.div[_.div(id='run-time-chart', class_='google-chart',
                                   style=f'height:{100 + 15 * number_of_child_runs}px')[' '],
                             _.script[f'''
drawRunTimeChart('run-time-chart', '{path}', {json.dumps(rows)});
    ''']])
        else:
            return str(_.i(style='color:#888')['Not enough data'])



"""Functions for keeping track of whether a list of files changed since the last pipeline run"""
import datetime
import hashlib
import pathlib

import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base

import mara_db.postgresql
from .. import config

Base = declarative_base()


class FileDependency(Base):
    """A combined hash of the content of a list of files"""
    __tablename__ = 'data_integration_file_dependency'

    node_path = sqlalchemy.Column(sqlalchemy.ARRAY(sqlalchemy.TEXT), primary_key=True)
    dependency_type = sqlalchemy.Column(sqlalchemy.String, primary_key=True)
    hash = sqlalchemy.Column(sqlalchemy.String)
    timestamp = sqlalchemy.Column(sqlalchemy.TIMESTAMP(timezone=True))


def update(node_path: [str], dependency_type: str, pipeline_base_path: str, file_dependencies: [str]):
    """
    Stores the combined hash of a list of files

    Args:
        node_path: The path of the node that depends on the files
        dependency_type: An arbitrary string that allows to distinguish between multiple dependencies of a node
        pipeline_base_path: The base directory of the pipeline
        file_dependencies: A list of file names relative to pipeline_base_path
    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f"""
INSERT INTO data_integration_file_dependency (node_path, dependency_type, hash, timestamp)
VALUES ({'%s,%s,%s,%s'})
ON CONFLICT (node_path, dependency_type)
DO UPDATE SET timestamp = EXCLUDED.timestamp, hash = EXCLUDED.hash
    """, (node_path, dependency_type, hash(pipeline_base_path, file_dependencies), datetime.datetime.utcnow()))

def delete(node_path: [str], dependency_type: str):
    """
    Delets the combined hash of a list of files for that node and dependency type

    Args:
        node_path: The path of the node that depends on the files
        dependency_type: An arbitrary string that allows to distinguish between multiple dependencies of a node
    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f"""
DELETE FROM data_integration_file_dependency
WHERE node_path = {'%s'} AND dependency_type = {'%s'}
    """, (node_path, dependency_type))


def is_modified(node_path: [str], dependency_type: str, pipeline_base_path: str, file_dependencies: [str]):
    """
    Checks whether a list of files have been modified since the last pipeline run

    Args:
        node_path: The path of the node that depends on the files
        dependency_type: An arbitrary string that allows to distinguish between multiple dependencies of a node
        pipeline_base_path: The base directory of the pipeline
        file_dependencies: A list of file names relative to pipeline_base_path

    Returns: True when at least one of the files was modified

    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute("""
SELECT TRUE
FROM data_integration_file_dependency
WHERE node_path=%s AND dependency_type=%s AND hash=%s """,
                       (node_path, dependency_type, hash(pipeline_base_path, file_dependencies)))
        return False if cursor.fetchone() else True


def hash(pipeline_base_path: pathlib.Path, file_dependencies: [str]) -> str:
    """
    Creates a combined hash of the content of a list of files

    Args:
        pipeline_base_path: The base directory of the pipeline
        file_dependencies: A list of file names relative to pipeline_base_path

    Returns: a combined content hash
    """
    hash = str(config.first_date()) + ' ' + str(config.last_date())
    for file_dependency in file_dependencies:
        hash += ' ' + hashlib.md5((pipeline_base_path / pathlib.Path(file_dependency)).read_text().encode()).hexdigest()
    return hash



"""Visualization of pipeline nodes"""

import functools
import json

import flask

from mara_page import _, bootstrap, html, response, acl
from . import views, last_runs, dependency_graph, run_time_chart
from .. import pipelines, config


@views.blueprint.route('/<path:path>')
@views.blueprint.route('', defaults={'path': ''})
def node_page(path: str):
    """Creates a node visualization page including title, action buttons, etc."""
    node, found = pipelines.find_node(path.split('/'))
    if not found and node:
        return flask.redirect(views.node_url(node), 302)
    elif not node:
        flask.abort(404, f'Node "{path}" not found')

    title = [node.__class__.__name__, ' ',
             [[_.a(href=views.node_url(parent))[parent.id], ' / '] for parent in node.parents()[1:-1]],
             node.id] if node.parent else 'Data Integration'
    return response.Response(
        title=title,
        action_buttons=action_buttons(node) if config.allow_run_from_web_ui() else [],
        html=[_.script['''
var nodePage = null;
document.addEventListener('DOMContentLoaded', function() {
     nodePage = NodePage("''' + flask.url_for('data_integration.node_page', path='') + '''", '''
                       + json.dumps(node.path()) + ''');
});'''],
              dependency_graph.card(node),
              run_time_chart.card(node),
              node_content(node),
              last_runs.card(node)],
        js_files=['https://www.gstatic.com/charts/loader.js',
                  flask.url_for('data_integration.static', filename='node-page.js'),
                  flask.url_for('data_integration.static', filename='utils.js'),
                  flask.url_for('data_integration.static', filename='run-time-chart.js'),
                  flask.url_for('data_integration.static', filename='system-stats-chart.js'),
                  flask.url_for('data_integration.static', filename='timeline-chart.js'),
                  flask.url_for('data_integration.static', filename='kolorwheel.js')],
        css_files=[flask.url_for('data_integration.static', filename='common.css'),
                   flask.url_for('data_integration.static', filename='node-page.css'),
                   flask.url_for('data_integration.static', filename='timeline-chart.css')])


@functools.singledispatch
def node_content(_: object) -> str:
    """Renders the node class specific part of a node page"""
    raise NotImplementedError()


@node_content.register(pipelines.Pipeline)
def __(pipeline: pipelines.Pipeline):
    return bootstrap.card(
        header_left='Nodes',
        header_right=[
            bootstrap.button(
                id='run-with-upstreams-button', label='Run with upstreams', icon='play',
                url=flask.url_for('data_integration.run_page', path=pipeline.url_path(), with_upstreams=True),
                title=f'Run selected nodes with all their upstreams in pipeline "{pipeline.id}"'),
            '&nbsp;&nbsp;&nbsp;&nbsp;',
            bootstrap.button(
                id='run-button', label='Run ', icon='play',
                url=flask.url_for('data_integration.run_page', path=pipeline.url_path(),
                                  with_upstreams=False),

                title='Run selected nodes')

        ] if config.allow_run_from_web_ui() else [],
        body=html.asynchronous_content(
            url=flask.url_for('data_integration.pipeline_children_table', path=pipeline.url_path())))


@node_content.register(pipelines.Task)
def __(task: pipelines.Task):
    if not acl.current_user_has_permission(views.acl_resource):
        return bootstrap.card(header_left='Commands', body=acl.inline_permission_denied_message())
    else:
        commands_card = bootstrap.card(
            header_left='Commands',
            fixed_header_height=True,
            sections=[_render_command(command) for command in task.commands])
        if task.max_retries:
            return [bootstrap.card(header_left=f'Max retries: {task.max_retries}'), commands_card]
        else:
            return commands_card


@node_content.register(pipelines.ParallelTask)
def __(task: pipelines.ParallelTask):
    if not acl.current_user_has_permission(views.acl_resource):
        return bootstrap.card(header_left=acl.inline_permission_denied_message())
    else:
        return [
            bootstrap.card(
                header_left='Commands before',
                fixed_header_height=True,
                sections=[_render_command(command) for command in task.commands_before]
            ) if task.commands_before else '',

            bootstrap.card(header_left='Sub task creation',
                           body=bootstrap.table([], [_.tr[_.td[_.div[section]], _.td(style='width:90%')[content]]
                                                     for section, content in task.html_doc_items()])),

            bootstrap.card(
                header_left='Commands after',
                fixed_header_height=True,
                sections=[_render_command(command) for command in task.commands_after]
            ) if task.commands_after else '']


def _render_command(command: pipelines.Command):
    """Creates a html documentation for a command"""
    from mara_page.xml import render

    def __mask_passwords(content):
        masks = config.password_masks()
        if masks:
            content = ''.join(render(content))
            for mask in masks:
                content = content.replace(mask, "***")
        return content

    try:
        doc = bootstrap.table([], [_.tr[_.td[_.div[section]], _.td(style='width:90%')[__mask_passwords(content)]]
                                   for section, content in command.html_doc_items()])
    except Exception as e:
        import traceback
        doc = [_.p(style='color:red')[_.i['Error in rendering command documentation']],
               _.pre(style='color:red')[traceback.format_exc()]]

    return [_.p[_.b[command.__class__.__name__]], doc]


@functools.singledispatch
def action_buttons(node: pipelines.Node):
    """The action buttons to be displayed on a node page"""
    path = node.path()
    return [
        response.ActionButton(
            action=flask.url_for('data_integration.run_page', path='/'.join(path[:-1]),
                                 with_upstreams=True, ids=path[-1]),
            label='Run with upstreams', icon='play',
            title=f'Run the task and all its upstreams in the pipeline "{node.parent.id}"'),
        response.ActionButton(
            action=flask.url_for('data_integration.run_page', path='/'.join(path[:-1]),
                                 with_upstreams=False, ids=path[-1]),
            label='Run', icon='play',
            title=f'Run only this task, without upstreams')]


@action_buttons.register(pipelines.Pipeline)
def __(pipeline: pipelines.Pipeline):
    return [response.ActionButton(action=flask.url_for('data_integration.run_page', path=pipeline.url_path(),
                                                       with_upstreams=False),
                                  label='Run', icon='play',
                                  title='Run the pipeline')]


@views.blueprint.route('/<path:path>/pipeline-children-table')
@views.blueprint.route('/pipeline-children-table', defaults={'path': ''})
@acl.require_permission(views.acl_resource, do_abort=False)
def pipeline_children_table(path: str):
    """Creates a table that documents all child nodes of a table"""
    from ..logging import node_cost

    pipeline, __ = pipelines.find_node(path.split('/'))
    assert (isinstance(pipeline, pipelines.Pipeline))

    node_durations_and_run_times = node_cost.node_durations_and_run_times(pipeline)

    rows = []
    for node in pipeline.nodes.values():
        [avg_duration, avg_run_time] = node_durations_and_run_times.get(tuple(node.path()), ['', ''])

        rows.append(
            _.tr[_.td[_.a(href=views.node_url(node))[node.id.replace('_', '_<wbr>')]],
                 _.td[node.description],
                 _.td[views.format_labels(node)],
                 _.td[node_cost.format_duration(avg_duration)],
                 _.td(style='color:#bbb' if avg_duration == avg_run_time else '')[
                     node_cost.format_duration(avg_run_time)],
                 _.td[node_cost.format_duration(
                     node_cost.compute_cost(node, node_durations_and_run_times))],
                 _.td[(_.input(class_='pipeline-node-checkbox', type='checkbox',
                               value=node.id, name='ids[]', onchange='runButtons.update()')
                       if config.allow_run_from_web_ui() else '')]])

    return \
        str(_.script['var runButtons = new PipelineRunButtons();']) \
        + str(bootstrap.table(['ID', 'Description', '', 'Avg duration', 'Avg run time', 'Cost', ''], rows)) \
        + str(_.script['floatMaraTableHeaders();'])



"""Slack notifications for failed node runs"""

from data_integration import config
from data_integration.logging import pipeline_events
from data_integration import events


class Slack(events.EventHandler):
    node_output: {tuple: {bool: [events.Event]}} = None

    def handle_event(self, event: events.Event):
        """
        Send the output of a node to Slack when the node failed.
        Args:
            event: The current event of interest
        """
        import requests

        if isinstance(event, pipeline_events.Output):
            key = tuple(event.node_path)

            if not self.node_output:
                self.node_output = {}

            if not key in self.node_output:
                self.node_output[key] = {True: [], False: []}

            self.node_output[key][event.is_error].append(event)

        elif isinstance(event, pipeline_events.NodeFinished):
            key = tuple(event.node_path)
            if not event.succeeded and event.is_pipeline is False:

                message = {'text': '\n:baby_chick: Ooops, a hiccup in '
                                   + '_ <' + config.base_url() + '/' + '/'.join(event.node_path)
                                   + ' | ' + '/'.join(event.node_path) + ' > _',
                           'attachments': []}

                if (self.node_output[key][False]):
                    message['attachments'].append({'text': self.format_output(self.node_output[key][False]),
                                                   'mrkdwn_in': ['text']})

                if (self.node_output[key][True]):
                    message['attachments'].append({'text': self.format_output(self.node_output[key][True]),
                                                   'color': '#eb4d5c', 'mrkdwn_in': ['text']})

                response = requests.post('https://hooks.slack.com/services/' + config.slack_token(), json=message)

                if response.status_code != 200:
                    raise ValueError(
                        'Request to slack returned an error %s. The response is:\n%s' % (
                            response.status_code, response.text)
                    )
        elif isinstance(event, pipeline_events.RunStarted):
            # default handler only handles interactively started runs
            if event.interactively_started:
                message = f':hatching_chick: *{event.user}* manually triggered run of '
                message += ('pipeline <' + config.base_url() + '/' + '/'.join(event.node_path) + '|'
                            + '/'.join(event.node_path) + ' >' if not event.is_root_pipeline else 'root pipeline')

                if event.node_ids:
                    message += ', nodes ' + ', '.join([f'`{id_}`' for id_ in event.node_ids])

                requests.post('https://hooks.slack.com/services/' + config.slack_token(), json={'text': message})
        elif isinstance(event, pipeline_events.RunFinished):
            # default handler only handles interactively started runs
            if event.interactively_started:
                if event.succeeded:
                    msg = ':hatched_chick: succeeded'
                else:
                    msg = ':baby_chick: failed'
                requests.post('https://hooks.slack.com/services/' + config.slack_token(),
                              json={'text': msg})

    def format_output(self, output_events: [pipeline_events.Output]):
        output, last_format = '', ''
        for event in output_events:
            if event.format == pipeline_events.Output.Format.VERBATIM:
                if last_format == event.format:
                    # append new verbatim line to the already initialized verbatim output
                    output = output[0:-3] + '\n' + event.message + '```'
                else:
                    output += '\n' + '```' + event.message + '```'
            elif event.format == pipeline_events.Output.Format.ITALICS:
                for line in event.message.splitlines():
                    output += '\n _ ' + str(line) + ' _ '
            else:
                output = '\n' + event.message

            last_format = event.format
        return output



"""Commands for running python functions and scripts"""

import inspect
import shlex
import sys
import json
from html import escape
from typing import Union, Callable, List

from mara_page import html, _
from .. import pipelines


class RunFunction(pipelines.Command):
    def __init__(self, function: Callable = None, args: [str] = None) -> None:
        """
        Runs an arbitrary python function

        Args:
            function: The parameterless function to run
            args: A list of arguments to be passed to the script
        Note:
            if you want to pass arguments, then use a lambda function
        """
        self.function = function
        self.args = args or []

    def run(self) -> bool:
        return self.function(*self.args)

    def html_doc_items(self) -> [(str, str)]:
        return [('function', _.pre[escape(str(self.function))]),
                ('args', _.tt[repr(self.args)]),
                (_.i['implementation'], html.highlight_syntax(inspect.getsource(self.function), 'python'))]


class ExecutePython(pipelines.Command):
    def __init__(self, file_name: Union[Callable, str],
                 args: Union[Callable, List[str]] = None) -> None:
        """
        Runs a python script in a separate interpreter process

        Args:
            file_name: the path of the file to run, relative to the pipeline directory
            args: A list of arguments to be passed to the script
        """
        self._file_name = file_name
        self._args = args or []

    @property
    def file_name(self):
        return self._file_name() if callable(self._file_name) else self._file_name

    @property
    def args(self):
        return self._args() if callable(self._args) else self._args

    def shell_command(self):
        return f'{shlex.quote(sys.executable)} -u "{self.parent.parent.base_path() / self.file_name}" {" ".join(map(str, self.args))}'

    def html_doc_items(self):
        path = self.parent.parent.base_path() / self.file_name
        return [
            ('file name', _.i[self.file_name]),
            ('args', _.tt[json.dumps(self.args)]),
            (_.i['content'], html.highlight_syntax(path.read_text().strip('\n') if path.exists() else '',
                                                   'python'))
        ]



"""Configuration of data integration pipelines and how to run them"""

import datetime
import multiprocessing
import pathlib
import typing
import functools

from . import pipelines, events


def root_pipeline() -> 'pipelines.Pipeline':
    """A pipeline that contains all other pipelines of the project"""
    return pipelines.demo_pipeline()


def data_dir() -> str:
    """Where to find local data files"""
    return str(pathlib.Path('data').absolute())


def default_db_alias() -> str:
    """The alias of the database that should be used when not specified otherwise"""
    return 'dwh-etl'


def first_date() -> datetime.date:
    """Ignore data before this date"""
    return datetime.date(2000, 1, 1)


def last_date() -> datetime.date:
    """Ignore data after this date"""
    return datetime.date(3000, 1, 1)


def max_number_of_parallel_tasks():
    """How many tasks can run in parallel at maximum"""
    return multiprocessing.cpu_count()


def bash_command_string() -> str:
    """The command used for running a bash, should somehow include the `pipefail` option"""
    return '/usr/bin/env bash -o pipefail'


def system_statistics_collection_period() -> int:
    """How often should system statistics be collected in seconds"""
    return 1


def run_log_retention_in_days() -> int:
    """How many days to keep node run times, output logs and system statistics"""
    return 30


def allow_run_from_web_ui() -> bool:
    """When false, then it is not possible to run an ETL from the web UI"""
    return True


def base_url() -> str:
    """External url of flask app, for linking nodes in slack messages"""
    return 'http://127.0.0.1:5000/data-integration'


def slack_token() -> typing.Optional[str]:
    """
    When not None, then this slack webhook is notified of failed nodes.
    Slack channel's token (i.e. THISIS/ASLACK/TOCKEN) can be retrieved from the
    channel's app "Incoming WebHooks" configuration as part part of the Webhook URL
    """
    return None

@functools.lru_cache(maxsize=None)
def event_handlers() -> [events.EventHandler]:
    """User specific envent handlers (mainly to notify chat systems)"""
    configured_handlers = []
    if slack_token():
        from data_integration.notification import slack
        configured_handlers.append(slack.Slack())
    return configured_handlers


def password_masks() -> typing.List[str]:
    """Any passwords which should be masked in the UI or logs"""

    # If you don't want to show the passwords in the password dialog, use something like this
    # to hide them there as well:
    #
    # class MasksList(list):
    #     def __repr__(self):
    #         return f"<MasksList with {len(self)} elements>"

    return []



import datetime
import enum
import glob
import json
import math
import os.path
import pathlib
import re
from html import escape

import mara_db.config
import mara_db.dbs
import mara_db.postgresql
from mara_page import _, html

from .. import config, pipelines
from ..commands import python, sql, files
from ..incremental_processing import file_dependencies as _file_dependencies
from ..incremental_processing import processed_files as _processed_files
from ..logging import logger


class ReadMode(enum.EnumMeta):
    """A mode for specifying which files from a list of files to load"""
    ALL = 'all'  # load all files that match the pattern
    ONLY_LATEST = 'only_latest'  # load only the latest matching file (requires date-regex)
    ONLY_NEW = 'only_new'  # load only files that have not been loaded yet
    ONLY_CHANGED = 'only_changed'  # load all files that were modified since the last run (or that are new)
    ONLY_NEW_EXCEPT_LATEST = 'only_new_except_latest'  # load only files that have not been loaded yet and not the last one


class _ParallelRead(pipelines.ParallelTask):
    def __init__(self, id: str, description: str, file_pattern: str, read_mode: ReadMode, target_table: str,
                 max_number_of_parallel_tasks: int = None, file_dependencies: [str] = None, date_regex: str = None,
                 partition_target_table_by_day_id: bool = False, truncate_partitions: bool = False,
                 commands_before: [pipelines.Command] = None, commands_after: [pipelines.Command] = None,
                 db_alias: str = None, timezone: str = None) -> None:
        pipelines.ParallelTask.__init__(self, id=id, description=description,
                                        max_number_of_parallel_tasks=max_number_of_parallel_tasks,
                                        commands_before=commands_before, commands_after=commands_after)
        self.file_pattern = file_pattern
        self.read_mode = read_mode
        self.date_regex = date_regex
        self.file_dependencies = file_dependencies or []
        self.partition_target_table_by_day_id = partition_target_table_by_day_id
        if self.partition_target_table_by_day_id:
            assert date_regex

        self.truncate_partitions = truncate_partitions
        if self.truncate_partitions:
            assert partition_target_table_by_day_id

        self.target_table = target_table
        self._db_alias = db_alias
        self.timezone = timezone

    @property
    def db_alias(self):
        return self._db_alias or config.default_db_alias()

    def add_parallel_tasks(self, sub_pipeline: 'pipelines.Pipeline') -> None:
        import more_itertools

        files = []  # A list of (file_name, date_or_file_name) tuples
        data_dir = config.data_dir()
        first_date = config.first_date()

        for file in glob.iglob(str(pathlib.Path(data_dir, self.file_pattern))):
            file = str(pathlib.Path(file).relative_to(pathlib.Path(data_dir)))
            if self.date_regex:
                match = re.match(self.date_regex, file)
                if not match:
                    raise Exception(f'file name "{file}" \ndoes not match date regex "{self.date_regex}"')
                date = datetime.date(*[int(group) for group in match.groups()])
                if date >= first_date:
                    files.append((file, date))
            else:
                files.append((file, file))

        # sort by date when regex provided or by filename otherwise
        files.sort(key=lambda x: x[1], reverse=True)

        # remove latest file when requested
        if self.read_mode == ReadMode.ONLY_NEW_EXCEPT_LATEST:
            files = files[1:]

        # take only latest file when requested
        if files and len(files) > 0 and self.read_mode == ReadMode.ONLY_LATEST:
            files = files[:1]

        # for incremental loading, determine which files already have been processed
        # reprocess all when file dependencies changed
        if (self.read_mode not in (ReadMode.ALL, ReadMode.ONLY_LATEST)
                and (not self.file_dependencies
                     or not _file_dependencies.is_modified(self.path(), 'ParallelReadFile', self.parent.base_path(),
                                                           self.file_dependencies))):
            processed_files = _processed_files.already_processed_files(self.path())

            files = [x for x in files
                     if x[0] not in processed_files  # everything not yet read
                     or (self.read_mode == ReadMode.ONLY_CHANGED  # everything modified
                         and self._last_modification_timestamp(x[0]) > processed_files[x[0]])]

        if not files:
            logger.log('No newer files', format=logger.Format.ITALICS)
            return

        if self.read_mode != ReadMode.ALL and self.file_dependencies:
            def update_file_dependencies():
                _file_dependencies.update(self.path(), 'ParallelReadFile', self.parent.base_path(),
                                          self.file_dependencies)
                return True

            sub_pipeline.final_node.commands.append(python.RunFunction(update_file_dependencies))

        chunk_size = math.ceil(len(files) / (2 * config.max_number_of_parallel_tasks()))

        if self.partition_target_table_by_day_id:
            if not isinstance(mara_db.dbs.db(self.db_alias), mara_db.dbs.PostgreSQLDB):
                raise NotImplementedError(
                    f'Partitioning by day_id has only been implemented for postgresql so far, \n'
                    f'not for {mara_db.postgresql.engine(self.db_alias).name}')
            files_per_day = {}
            for (file, date) in files:
                if date in files_per_day:
                    files_per_day[date].append(file)
                else:
                    files_per_day[date] = [file]

            sql_statements = []
            for date in files_per_day.keys():
                sql_statements.append(f'CREATE TABLE IF NOT EXISTS {self.target_table}_{date.strftime("%Y%m%d")}'
                                      + f' PARTITION OF {self.target_table} FOR VALUES IN ({date.strftime("%Y%m%d")});')

                if self.truncate_partitions:
                    sql_statements.append(f'TRUNCATE {self.target_table}_{date.strftime("%Y%m%d")};')

            create_partitions_task = pipelines.Task(
                id='create_partitions',
                description='Creates required target table partitions',
                commands=[sql.ExecuteSQL(sql_statement='\n'.join(slice), echo_queries=False, db_alias=self.db_alias)
                          for slice in more_itertools.sliced(sql_statements, 50)])

            sub_pipeline.add(create_partitions_task)

            for n, chunk in enumerate(more_itertools.chunked(files_per_day.items(), chunk_size)):
                task = pipelines.Task(id=str(n), description='Reads a portion of the files')
                for (day, files) in chunk:
                    target_table = self.target_table + '_' + day.strftime("%Y%m%d")
                    for file in files:
                        task.add_commands(self.parallel_commands(file))
                    task.add_command(sql.ExecuteSQL(sql_statement=f'ANALYZE {target_table}'))
                sub_pipeline.add(task, ['create_partitions'])
        else:
            for n, chunk in enumerate(more_itertools.chunked(files, chunk_size)):
                sub_pipeline.add(
                    pipelines.Task(id=str(n), description=f'Reads {len(chunk)} files',
                                   commands=sum([self.parallel_commands(x[0]) for x in chunk], [])))

    def parallel_commands(self, file_name: str) -> [pipelines.Command]:
        return [self.read_command(file_name)] + (
            [python.RunFunction(function=lambda: _processed_files.track_processed_file(
                self.path(), file_name, self._last_modification_timestamp(file_name)))]
            if self.read_mode != ReadMode.ALL else [])

    def read_command(self) -> pipelines.Command:
        raise NotImplementedError

    def _last_modification_timestamp(self, file_name):
        return datetime.datetime.fromtimestamp(
            os.path.getmtime(pathlib.Path(config.data_dir()) / file_name)).astimezone()


class ParallelReadFile(_ParallelRead):
    def __init__(self, id: str, description: str, file_pattern: str, read_mode: ReadMode,
                 compression: files.Compression, target_table: str, file_dependencies: [str] = None,
                 date_regex: str = None, partition_target_table_by_day_id: bool = False,
                 truncate_partitions: bool = False,
                 commands_before: [pipelines.Command] = None, commands_after: [pipelines.Command] = None,
                 mapper_script_file_name: str = None, make_unique: bool = False, db_alias: str = None,
                 delimiter_char: str = None, quote_char: str = None, null_value_string: str = None,
                 skip_header: bool = None, csv_format: bool = False,
                 timezone: str = None, max_number_of_parallel_tasks: int = None) -> None:
        _ParallelRead.__init__(self, id=id, description=description, file_pattern=file_pattern,
                               read_mode=read_mode, target_table=target_table, file_dependencies=file_dependencies,
                               date_regex=date_regex, partition_target_table_by_day_id=partition_target_table_by_day_id,
                               truncate_partitions=truncate_partitions,
                               commands_before=commands_before, commands_after=commands_after,
                               db_alias=db_alias, timezone=timezone,
                               max_number_of_parallel_tasks=max_number_of_parallel_tasks)
        self.compression = compression
        self.mapper_script_file_name = mapper_script_file_name or ''
        self.make_unique = make_unique
        self.delimiter_char = delimiter_char
        self.quote_char = quote_char
        self.skip_header = skip_header
        self.csv_format = csv_format
        self.null_value_string = null_value_string

    def read_command(self, file_name: str) -> pipelines.Command:
        return files.ReadFile(file_name=file_name, compression=self.compression, target_table=self.target_table,
                              mapper_script_file_name=self.mapper_script_file_name, make_unique=self.make_unique,
                              db_alias=self.db_alias, delimiter_char=self.delimiter_char, skip_header=self.skip_header,
                              quote_char=self.quote_char, null_value_string=self.null_value_string,
                              csv_format=self.csv_format, timezone=self.timezone)

    def html_doc_items(self) -> [(str, str)]:
        path = self.parent.base_path() / self.mapper_script_file_name if self.mapper_script_file_name else ''
        return [('file pattern', _.i[self.file_pattern]),
                ('compression', _.tt[self.compression]),
                ('read mode', _.tt[self.read_mode]),
                ('date regex', _.tt[escape(self.date_regex)] if self.date_regex else None),
                ('file dependencies', [_.i[dependency, _.br] for dependency in self.file_dependencies]),
                ('mapper script file name', _.i[self.mapper_script_file_name]),
                (_.i['mapper script'], html.highlight_syntax(path.read_text().strip('\n')
                                                             if self.mapper_script_file_name and path.exists()
                                                             else '', 'python')),
                ('make unique', _.tt[repr(self.make_unique)]),
                ('skip header', _.tt[self.skip_header]),
                ('target_table', _.tt[self.target_table]),
                ('db alias', _.tt[self.db_alias]),
                ('partion target table by day_id', _.tt[self.partition_target_table_by_day_id]),
                ('truncate partitions', _.tt[self.truncate_partitions]),
                ('sql delimiter char',
                 _.tt[json.dumps(self.delimiter_char) if self.delimiter_char != None else None]),
                ('quote char', _.tt[json.dumps(self.quote_char) if self.quote_char != None else None]),
                ('null value string',
                 _.tt[json.dumps(self.null_value_string) if self.null_value_string != None else None]),
                ('time zone', _.tt[self.timezone])]


class ParallelReadSqlite(_ParallelRead):
    def __init__(self, id: str, description: str, file_pattern: str, read_mode: ReadMode, sql_file_name: str,
                 target_table: str, file_dependencies: [str] = None, date_regex: str = None,
                 partition_target_table_by_day_id: bool = False, truncate_partitions: bool = False,
                 commands_before: [pipelines.Command] = None, commands_after: [pipelines.Command] = None,
                 db_alias: str = None, timezone=None, max_number_of_parallel_tasks: int = None) -> None:
        _ParallelRead.__init__(self, id=id, description=description, file_pattern=file_pattern,
                               read_mode=read_mode, target_table=target_table, file_dependencies=file_dependencies,
                               date_regex=date_regex, partition_target_table_by_day_id=partition_target_table_by_day_id,
                               truncate_partitions=truncate_partitions,
                               commands_before=commands_before, commands_after=commands_after, db_alias=db_alias,
                               timezone=timezone, max_number_of_parallel_tasks=max_number_of_parallel_tasks)
        self.sql_file_name = sql_file_name

    def read_command(self, file_name: str) -> [pipelines.Command]:
        return files.ReadSQLite(sqlite_file_name=file_name, sql_file_name=self.sql_file_name,
                                target_table=self.target_table, db_alias=self.db_alias, timezone=self.timezone)

    def sql_file_path(self):
        return self.parent.base_path() / self.sql_file_name

    def html_doc_items(self) -> [(str, str)]:
        path = self.sql_file_path()
        return [('file pattern', _.i[self.file_pattern]),
                ('read mode', _.tt[self.read_mode]),
                ('date regex', _.tt[escape(self.date_regex)] if self.date_regex else None),
                ('file dependencies', [_.i[dependency, _.br] for dependency in self.file_dependencies]),
                ('query file name', _.i[self.sql_file_name]),
                (_.i['query'], html.highlight_syntax(path.read_text().strip('\n')
                                                     if self.sql_file_name and path.exists()
                                                     else '', 'sql')),
                ('target_table', _.tt[self.target_table]),
                ('db alias', _.tt[self.db_alias]),
                ('partion target table by day_id', _.tt[self.partition_target_table_by_day_id]),
                ('truncate partitions', _.tt[self.truncate_partitions]),
                ('time zone', _.tt[self.timezone])]



"""Computation of node duration, run time and cost"""

import functools
import math

import mara_db.config
import mara_db.postgresql
from .. import pipelines


def node_durations_and_run_times(node: pipelines.Node) -> {tuple: [float, float]}:
    """
    Returns for children of `node` the average duration and run time (sum of average duration of all leaf nodes)

    Args:
        node: The parent node

    Returns:
        A dictionary of {(node_path,): [avg_duration, avg_run_time]} entries for all children of `node`

    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f"""
WITH child_nodes AS
       (SELECT node.node_path [ 0 : {'%(level)s'} + 1]                     AS node_path,
               avg(CASE -- direct children
                     WHEN array_length(node.node_path, 1) = {'%(level)s'} + 1
                       THEN extract(EPOCH FROM end_time - start_time) END) AS avg_duration,
               avg(CASE -- all children except pipelines
                     WHEN NOT is_pipeline
                       THEN extract(EPOCH FROM end_time - start_time) END) AS avg_run_time
        FROM data_integration_node_run node
        WHERE node.node_path [ 0 : {'%(level)s'}] = %(path)s
          AND array_length(node.node_path, 1) > {'%(level)s'}
        GROUP BY node.node_path)
SELECT node_path,
       round(avg(avg_duration)::NUMERIC, 1) AS avg_duration,
       round(sum(avg_run_time)::NUMERIC, 1) AS avg_run_time
FROM child_nodes
GROUP BY node_path;""", {'path': node.path(), 'level': len(node.path())})

        return {tuple(row[0]): row[1:] for row in cursor.fetchall()}


def compute_cost(node: pipelines.Node, node_durations_and_run_times: {tuple: [float, float]}) -> float:
    """
    Computes the cost of a node as maximum cumulative run time of a node and all its downstreams.
    Stores the result in `node` and also returns it

    Args:
        node: The node for which to compute the cost
        node_durations_and_run_times: Duration and run time information as computed by
                                      the `node_durations_and_run_times` function
    """
    path = tuple(node.path())
    if node.cost is None:
        node.cost = (
                max([compute_cost(downstream, node_durations_and_run_times)
                     for downstream in node.downstreams] or [0])
                + (node_durations_and_run_times.get(path, [0, 0])[1] or 0))

    return node.cost


@functools.lru_cache(maxsize=None)
def format_duration(duration: float) -> str:
    """
    Formats a duration in human readable form

    Args:
        duration: A duration in seconds

    Returns:
        The duration as a string

    Examples:
        >>>> print(format_duration(0.12))
        0.12s

        >>>> print(format_duration(5.6))
        5.6s

        >>>> print(format_duration(70.2))
        1:10m

        >>>> print(format_duration(4000))
        1:06h
    """
    if duration in [None, '']:
        return ''

    hours = math.floor(duration / 3600)
    duration -= 3600 * hours

    minutes = math.floor(duration / 60)
    duration -= 60 * minutes

    seconds = math.floor(duration)

    if hours:
        return f"{hours}:{str(minutes).rjust(2, '0')[:2]}h"
    elif minutes:
        return f"{minutes}:{str(seconds).rjust(2, '0')[:2]}m"
    else:
        return f"{round(duration, 1)}s"



"""Command line interface for running data pipelines"""

import sys

import click

from .. import config, pipelines


def run_pipeline(pipeline: pipelines.Pipeline, nodes: {pipelines.Node} = None,
                 with_upstreams: bool = False,
                 interactively_started: bool = False) -> bool:
    """
    Runs a pipeline or parts of it with output printed to stdout
    Args:
        pipeline: The pipeline to run
        nodes: A list of pipeline children that should run
        with_upstreams: When true and `nodes` are provided, then all upstreams of `nodes` in `pipeline` are also run
        interactively_started: Whether or not this run was started interactively, passed on in RunStarted and
                               RunFinished events.
    Return:
        True when the pipeline run succeeded
    """
    from ..logging import logger, pipeline_events
    from .. import execution

    succeeded = True
    for event in execution.run_pipeline(pipeline, nodes, with_upstreams, interactively_started=interactively_started):
        if isinstance(event, pipeline_events.Output):
            print(f'\033[36m{" / ".join(event.node_path)}{":" if event.node_path else ""}\033[0m '
                  + {logger.Format.STANDARD: '\033[01m',
                     logger.Format.ITALICS: '\033[02m',
                     logger.Format.VERBATIM: ''}[event.format]
                  + ('\033[91m' if event.is_error else '') + event.message + '\033[0m')
        elif isinstance(event, pipeline_events.RunFinished):
            if not event.succeeded:
                succeeded = False

    return succeeded


@click.command()
@click.option('--path', default='',
              help='The parent ids of of the pipeline to run, separated by comma. Example: "pipeline-id,sub-pipeline-id".')
@click.option('--nodes',
              help='IDs of sub-nodes of the pipeline to run, separated by comma. When provided, then only these nodes are run. Example: "do-this, do-that".')
@click.option('--with_upstreams', default=False, is_flag=True,
              help='Also run all upstreams of --nodes within the pipeline.')
def run(path, nodes, with_upstreams):
    """Runs a pipeline or a sub-set of its nodes"""

    # the pipeline to run
    path = path.split(',')
    pipeline, found = pipelines.find_node(path)
    if not found:
        print(f'Pipeline {path} not found', file=sys.stderr)
        sys.exit(-1)
    if not isinstance(pipeline, pipelines.Pipeline):
        print(f'Node {path} is not a pipeline, but a {pipeline.__class__.__name__}', file=sys.stderr)
        sys.exit(-1)

    # a list of nodes to run selectively in the pipeline
    _nodes = set()
    for id in (nodes.split(',') if nodes else []):
        node = pipeline.nodes.get(id)
        if not node:
            print(f'Node "{id}" not found in pipeline {path}', file=sys.stderr)
            sys.exit(-1)
        else:
            _nodes.add(node)

    if not run_pipeline(pipeline, _nodes, with_upstreams, interactively_started=False):
        sys.exit(-1)


@click.command()
def run_interactively():
    """Select and run data pipelines"""
    from dialog import Dialog

    d = Dialog(dialog="dialog", autowidgetsize=True)  # see http://pythondialog.sourceforge.net/doc/widgets.html

    def run_pipeline_and_notify(pipeline: pipelines.Pipeline, nodes: {pipelines.Node} = None):
        if not run_pipeline(pipeline, nodes, interactively_started=True):
            sys.exit(-1)

    def menu(node: pipelines.Node):
        if isinstance(node, pipelines.Pipeline):

            code, choice = d.menu(
                text='Pipeline ' + '.'.join(node.path()) if node.parent else 'Root pipeline',
                choices=[('▶ ', 'Run'), ('>> ', 'Run selected')]
                        + [(child.id, '→' if isinstance(child, pipelines.Pipeline) else 'Run')
                           for child in node.nodes.values()])
            if code == d.CANCEL:
                return

            if choice == '▶ ':
                run_pipeline_and_notify(node)
            elif choice == '>> ':
                code, node_ids = d.checklist('Select sub-nodes to run. If you want to run all, then select none.',
                                             choices=[(node_id, '', False) for node_id in node.nodes.keys()])
                if code == d.OK:
                    run_pipeline_and_notify(node, {node.nodes[id] for id in node_ids})
            else:
                menu(node.nodes[choice])
            return
        else:
            run_pipeline_and_notify(pipeline=node.parent, nodes=[node])

    menu(config.root_pipeline())


@click.command()
@click.option('--path', default='',
              help='The parent ids of of the node to reset. Example: "pipeline-id,sub-pipeline-id".')
def reset_incremental_processing(path):
    """Reset status of incremental processing for a node"""
    from ..incremental_processing import reset

    path = path.split(',') if path else []
    node, found = pipelines.find_node(path)
    if not found:
        print(f'Node {path} not found', file=sys.stderr)
        sys.exit(-1)
    reset.reset_incremental_processing(path)



"""UI for running pipelines from the web browser"""

import json

import flask

from mara_page import _, bootstrap, response, acl
from . import views
from .. import pipelines, config


@views.blueprint.route('/run-', defaults={'path': '', 'with_upstreams': False, 'ids': None})
@views.blueprint.route('/run-/<path:ids>', defaults={'path': '', 'with_upstreams': False})
@views.blueprint.route('/run-with-upstreams', defaults={'path': '', 'with_upstreams': True, 'ids': None})
@views.blueprint.route('/run-with-upstreams/<path:ids>', defaults={'path': '', 'with_upstreams': True})
@views.blueprint.route('/<path:path>/run-', defaults={'with_upstreams': False, 'ids': None})
@views.blueprint.route('/<path:path>/run-/<path:ids>', defaults={'with_upstreams': False})
@views.blueprint.route('/<path:path>/run-with-upstreams', defaults={'with_upstreams': True, 'ids': None})
@views.blueprint.route('/<path:path>/run-with-upstreams/<path:ids>', defaults={'with_upstreams': True})
@acl.require_permission(views.acl_resource)
def run_page(path: str, with_upstreams: bool, ids: str):
    if not config.allow_run_from_web_ui():
        flask.abort(403, 'Running piplelines from web ui is disabled for this instance')

    # the pipeline to run
    pipeline, found = pipelines.find_node(path.split('/'))
    if not found:
        flask.abort(404, f'Pipeline "{path}" not found')
    assert (isinstance(pipeline, pipelines.Pipeline))

    # a list of nodes to run selectively in the pipeline
    nodes = []
    for id in (ids.split('/') if ids else []):
        node = pipeline.nodes.get(id)
        if not node:
            flask.abort(404, f'Node "{id}" not found in pipeline "{path}"')
        else:
            nodes.append(node)

    stream_url = flask.url_for('data_integration.do_run', path=path, with_upstreams=with_upstreams, ids=ids)

    title = ['Run ', 'with upstreams ' if with_upstreams else '',
             ' / '.join([str(_.a(href=views.node_url(parent))[parent.id]) for parent in pipeline.parents()[1:]])]
    if nodes:
        title += [' / [', ', '.join([str(_.a(href=views.node_url(node))[node.id]) for node in nodes]), ']']

    return response.Response(
        html=[
            _.script['''
document.addEventListener('DOMContentLoaded', function() {
     processRunEvents(''' + json.dumps(flask.url_for('data_integration.node_page', path='')) + ', '
                     + json.dumps(stream_url) + ', ' + json.dumps(pipeline.path()) + ''');
});'''],

            _.style['span.action-buttons > * {display:none}'],  # hide reload button until run finishes
            _.div(class_='row')[
                _.div(class_='col-lg-7')[
                    bootstrap.card(body=_.div(id='main-output-area', class_='run-output')[''])],
                _.div(class_='col-lg-5 scroll-container')[
                    bootstrap.card(header_left='Timeline',
                                   body=[_.div(id='system-stats-chart', class_='google-chart')[' '],
                                         _.div(id='timeline-chart')[' ']]),
                    _.div(id='failed-tasks-container')[''],
                    _.div(id='running-tasks-container')[''],
                    _.div(id='succeeded-tasks-container')[''],

                    bootstrap.card(id='card-template', header_left=' ', header_right=' ',
                                   body=[_.div(class_='run-output')['']])
                ]
            ]
        ],
        js_files=['https://www.gstatic.com/charts/loader.js',
                  flask.url_for('data_integration.static', filename='timeline-chart.js'),
                  flask.url_for('data_integration.static', filename='system-stats-chart.js'),
                  flask.url_for('data_integration.static', filename='utils.js'),
                  flask.url_for('data_integration.static', filename='run-page.js')],
        css_files=[flask.url_for('data_integration.static', filename='timeline-chart.css'),
                   flask.url_for('data_integration.static', filename='run-page.css'),
                   flask.url_for('data_integration.static', filename='common.css')],
        action_buttons=[response.ActionButton(action='javascript:location.reload()', label='Run again', icon='play',
                                              title='Run pipeline again with same parameters as before')],
        title=title,

    )


@views.blueprint.route('/do-run', defaults={'path': '', 'with_upstreams': False, 'ids': None})
@views.blueprint.route('/do-run/<path:ids>', defaults={'path': '', 'with_upstreams': False})
@views.blueprint.route('/do-run-with-upstreams', defaults={'path': '', 'with_upstreams': True, 'ids': None})
@views.blueprint.route('/do-run-with-upstreams/<path:ids>', defaults={'path': '', 'with_upstreams': True})
@views.blueprint.route('/<path:path>/do-run', defaults={'with_upstreams': False, 'ids': None})
@views.blueprint.route('/<path:path>/do-run/<path:ids>', defaults={'with_upstreams': False})
@views.blueprint.route('/<path:path>/do-run-with-upstreams', defaults={'with_upstreams': True, 'ids': None})
@views.blueprint.route('/<path:path>/do-run-with-upstreams/<path:ids>', defaults={'with_upstreams': True})
@acl.require_permission(views.acl_resource)
def do_run(path: str, with_upstreams: bool, ids: str):
    from .. import execution

    if not config.allow_run_from_web_ui():
        flask.abort(403, 'Running piplelines from web ui is disabled for this instance')
    pipeline, found = pipelines.find_node(path.split('/'))
    if not found:
        flask.abort(404, f'Pipeline "{path}" not found')

    nodes = {pipeline.nodes[id] for id in (ids.split('/') if ids else [])}

    def process_events():
        for event in execution.run_pipeline(pipeline, nodes, with_upstreams):
            yield f'event: {event.__class__.__name__}\ndata: ' + event.to_json() + '\n\n'

    return flask.Response(process_events(), mimetype="text/event-stream")



import copy
import pathlib
import re
import typing

from . import config


class Node():
    """Base class for pipeline elements"""

    def __init__(self, id: str, description: str, labels: {str: str} = None) -> None:
        if not re.match('^[a-z0-9_]+$', id):
            raise ValueError(f'Invalid id "{id}". Should only contain lowercase letters, numbers and "_".')
        self.id: str = id
        self.description: str = description
        self.labels: {str: str} = labels or {}

        self.upstreams: {'Node'} = set()
        self.downstreams: {'Node'} = set()

        self.parent: typing.Optional['Pipeline'] = None
        self.cost: typing.Optional[float] = None


    def parents(self):
        """Returns all parents of a node from top to bottom"""
        if self.parent:
            return self.parent.parents() + [self]
        else:
            return [self]

    def path(self) -> [str]:
        """Returns a list of ids that identify the node across all pipelines, from top to bottom"""
        return [node.id for node in self.parents()[1:]]

    def url_path(self) -> str:
        """Returns a uri fragment for referring to nodes"""
        return '/'.join(self.path()) or None

    def __repr__(self):
        return f'<{self.__class__.__name__} "{self.id}">'


class Command():
    """
    Base class for operations that can run inside a pipeline task

    Args:
        parent: The pipeline node that contains this command (needed for debug output)
    """
    parent: Node = None

    def run(self) -> bool:
        """
        Runs the command

        Returns:
            False on failure
        """
        from . import shell
        shell_command = self.shell_command()

        # logger.log(f'{config.bash_command_string()} -c {shlex.quote(shell_command)}', format=logger.Format.ITALICS)
        return shell.run_shell_command(shell_command)

    def shell_command(self):
        """A bash command string that that runs the command"""
        raise NotImplementedError()

    def node_path(self):
        """The path of the parent node"""
        return self.parent.path() if self.parent else ''

    def html_doc_items(self) -> [(str, str)]:
        """
        Things to display in the documentation of a command. Can contain html
        Example: `[('filename','/tmp/foo.txt'), ('max-retries', 15)]`
        """
        raise NotImplementedError()


class Task(Node):
    def __init__(self, id: str, description: str, commands: [Command] = None, max_retries: int = 0) -> None:
        super().__init__(id, description)
        self.commands = []
        self.max_retries = max_retries

        for command in commands or []:
            self.add_command(command)

    def add_command(self, command: Command, prepend=False):
        if prepend:
            self.commands.insert(0, command)
        else:
            self.commands.append(command)
        command.parent = self

    def add_commands(self, commands: [Command]):
        for command in commands:
            self.add_command(command)

    def run(self):
        for command in self.commands:
            if not command.run():
                return False
        return True


class ParallelTask(Node):
    def __init__(self, id: str, description: str, max_number_of_parallel_tasks: int = None,
                 commands_before: [Command] = None, commands_after: [Command] = None) -> None:
        super().__init__(id, description)
        self.commands_before = []
        for command in commands_before or []:
            self.add_command_before(command)
        self.commands_after = []
        for command in commands_after or []:
            self.add_command_after(command)
        self.max_number_of_parallel_tasks = max_number_of_parallel_tasks

    def add_command_before(self, command: Command):
        self.commands_before.append(command)
        command.parent = self

    def add_command_after(self, command: Command):
        self.commands_after.append(command)
        command.parent = self

    def add_parallel_tasks(self, sub_pipeline: 'Pipeline') -> None:
        pass

    def launch(self) -> 'Pipeline':
        sub_pipeline = Pipeline(self.id, description=f'Runs f{self.id} in parallel',
                                max_number_of_parallel_tasks=self.max_number_of_parallel_tasks)
        sub_pipeline.add_initial(Task(id='before', description='Runs commands-before', commands=self.commands_before))
        sub_pipeline.add_final(Task(id='after', description='Runs commands-after', commands=self.commands_after))

        self.add_parallel_tasks(sub_pipeline)

        return sub_pipeline

    def html_doc_items(self) -> [(str, str)]:
        """
        Things to display in the documentation of the parallel task. Can contain html.
        Example: `[('filename','/tmp/foo.txt'), ('max-retries', 15)]`
        """
        raise NotImplementedError


class Pipeline(Node):
    nodes: {str: Node} = None
    initial_node: Node = None
    final_node: Node = None

    def __init__(self, id: str,
                 description: str,
                 max_number_of_parallel_tasks: int = None,
                 base_path: pathlib.Path = None,
                 labels: {str: str} = None,
                 ignore_errors: bool = False,
                 force_run_all_children: bool = False) -> None:
        """
        A directed acyclic graph (DAG) of nodes with dependencies between them.

        Args:
            id: The id of the pipeline
            description: A short summary of what the pipeline is doing
            max_number_of_parallel_tasks: Only that many nodes of the pipeline will run in parallel
            base_path: The absolute path of the pipeline root, file names are relative to that
            labels: An arbitrary dictionary application specific tags, schemas and so on.
            ignore_errors: When true, then the pipeline execution will not fail when a child node fails
            force_run_all_children: When true, child nodes will run even when their upstreams failed
        """
        super().__init__(id, description, labels)
        self.nodes = {}
        self._base_path = base_path
        self.max_number_of_parallel_tasks = max_number_of_parallel_tasks
        self.force_run_all_children = force_run_all_children
        self.ignore_errors = ignore_errors

    def add(self, node: Node, upstreams: [typing.Union[str, Node]] = None) -> 'Pipeline':
        if node.id in self.nodes:
            raise ValueError(f'A node with id "{node.id}" already exists in pipeline "{self.id}"')

        self.nodes[node.id] = node
        node.parent = self

        for upstream in upstreams or []:
            self.add_dependency(upstream, node)

        if self.initial_node and not node.upstreams and self.initial_node != node:
            self.add_dependency(self.initial_node, node)

        if self.final_node and self.final_node != node:
            self.add_dependency(node, self.final_node)

        return self

    def remove(self, node: Node) -> None:
        """
        Removes a `node` from the pipeline and its dependencies.
        All `upstreams` of `node` will be connected to all `downstreams` of `node`.

        Args:
            node: The node to remove
        """

        for upstream in node.upstreams:
            for downstream in node.downstreams:
                self.add_dependency(upstream, downstream)

        for upstream in copy.copy(node.upstreams):
            self.remove_dependency(upstream, node)

        for downstream in copy.copy(node.downstreams):
            self.remove_dependency(node, downstream)

        if self.initial_node == node:
            self.initial_node = None

        if self.final_node == node:
            self.final_node = None

        del (self.nodes[node.id])

    def replace(self, node: Node, new_node) -> Node:
        """
        Replaces `node` with `new_node` while keeping dependencies intact
        Args:
            node: The node to replace
            new_node: The new node
        """
        for upstream in copy.copy(node.upstreams):
            self.add_dependency(upstream, new_node)

        for downstream in copy.copy(node.downstreams):
            self.add_dependency(new_node, downstream)

        self.remove(node)
        self.add(new_node)

    def add_dependency(self, upstream: typing.Union[Node, str], downstream: typing.Union[Node, str]):
        if isinstance(upstream, str):
            if not upstream in self.nodes:
                raise KeyError(f'Node "{upstream}" not found in pipeline "{self.id}"')
            upstream = self.nodes[upstream]

        if isinstance(downstream, str):
            if not downstream in self.nodes:
                raise KeyError(f'Node "{downstream}" not found in pipeline "{self.id}"')
            downstream = self.nodes[downstream]

        upstream.downstreams.add(downstream)
        downstream.upstreams.add(upstream)

        if self.final_node and self.final_node != downstream:
            self.remove_dependency(upstream, self.final_node)

        if self.initial_node and self.initial_node != upstream:
            self.remove_dependency(self.initial_node, downstream)

    def remove_dependency(self, upstream: Node, downstream: Node):
        upstream.downstreams.discard(downstream)
        downstream.upstreams.discard(upstream)

    def add_initial(self, node: Node) -> 'Pipeline':
        self.initial_node = node
        for downstream in self.nodes.values():
            if not downstream.upstreams and downstream != self.final_node:
                self.add_dependency(node, downstream)
        self.add(node)

    def add_final(self, node: Node) -> 'Pipeline':
        self.final_node = node
        for upstream in self.nodes.values():
            if not upstream.downstreams and upstream != self.initial_node:
                self.add_dependency(upstream, node)
        self.add(node)

    def base_path(self):
        return self._base_path or (self.parent.base_path() if self.parent else pathlib.Path('.'))


def find_node(path: [str]) -> (Node, bool):
    """
    Retrieves a node by the the path from its parents
    Args:
        path: The ids of the node and of all its parent, from top to bottom

    Returns:
        A tuple of the node and True if the node was found, or a the closest known parent node if and False otherwise
    """
    if not path or path == ['']:
        return config.root_pipeline(), True

    def _find_node(node: Node, path) -> Node:
        if len(path) == 0:
            return node, True
        else:
            if isinstance(node, Pipeline) and path[0] in node.nodes:
                return _find_node(node.nodes[path[0]], path[1::])
            else:
                return node, False

    return _find_node(config.root_pipeline(), path)


def demo_pipeline():
    """Returns a demo pipeline"""
    from .commands import bash, python
    pipeline = Pipeline(id='demo',
                        description='A small pipeline that demonstrates the interplay between pipelines, tasks and commands')

    pipeline.add(Task(id='ping_localhost', description='Pings localhost',
                      commands=[bash.RunBash('ping -c 3 localhost')]))

    sub_pipeline = Pipeline(id='sub_pipeline', description='Pings a number of hosts')

    for host in ['google', 'amazon', 'facebook']:
        sub_pipeline.add(Task(id=f'ping_{host}', description=f'Pings {host}',
                              commands=[bash.RunBash(f'ping -c 3 {host}.com'),
                                        python.RunFunction(lambda: 1)]))
    sub_pipeline.add_dependency('ping_amazon', 'ping_facebook')
    sub_pipeline.add(Task(id='ping_foo', description='Pings foo',
                          commands=[bash.RunBash('ping foo')]), ['ping_amazon'])

    pipeline.add(sub_pipeline, ['ping_localhost'])

    pipeline.add(Task(id='sleep', description='Sleeps for 2 seconds',
                      commands=[bash.RunBash('sleep 2')]), ['sub_pipeline'])

    return pipeline



"""Functions for keeping track whether an input file has already been 'processed' """

import datetime

import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base

import mara_db.config
import mara_db.dbs
import mara_db.postgresql

Base = declarative_base()


class ProcessedFile(Base):
    """A local file that has been 'processed' (e.g. has been read)"""
    __tablename__ = 'data_integration_processed_file'

    node_path = sqlalchemy.Column(sqlalchemy.ARRAY(sqlalchemy.Text), primary_key=True)
    file_name = sqlalchemy.Column(sqlalchemy.Text, primary_key=True)
    last_modified_timestamp = sqlalchemy.Column(sqlalchemy.TIMESTAMP(timezone=True))


def track_processed_file(node_path: str, file_name: str, last_modified_timestamp: datetime):
    """
    Records that a file has been 'processed' by a node

    Args:
        node_path: The path of the node that processed the file
        file_name: The name of the file that has been processed
        last_modified_timestamp: The time when the file was modified last

    Returns: True
    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f'''
INSERT INTO data_integration_processed_file (node_path, file_name, last_modified_timestamp) 
VALUES ({'%s,%s,%s'})
ON CONFLICT (node_path, file_name) 
DO UPDATE SET last_modified_timestamp = EXCLUDED.last_modified_timestamp
''', (node_path, file_name, last_modified_timestamp))
    return True


def already_processed_files(node_path: str) -> {str: datetime}:
    """
    Returns all files that already have been processed by a node
    Args:
        node_path: The path of the node that processed the file

    Returns:
        A mapping of file names to timestamps of last modification
    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f"""
SELECT file_name, last_modified_timestamp
FROM data_integration_processed_file WHERE node_path = {'%s'}
""", (node_path,))
        return {row[0]: row[1] for row in cursor.fetchall()}



import inspect
import re
import typing

from mara_page import _, html
from .. import config, pipelines
from ..commands import sql


class ParallelExecuteSQL(pipelines.ParallelTask, sql._SQLCommand):
    def __init__(self, id: str, description: str, parameter_function: typing.Callable, parameter_placeholders: [str],
                 max_number_of_parallel_tasks: int = None, sql_statement: str = None, file_name: str = None,
                 commands_before: [pipelines.Command] = None, commands_after: [pipelines.Command] = None,
                 db_alias: str = None, echo_queries: bool = True, timezone: str = None,
                 replace: {str: str} = None) -> None:

        if (not (sql_statement or file_name)) or (sql_statement and file_name):
            raise ValueError('Please provide either sql_statement or file_name (but not both)')

        pipelines.ParallelTask.__init__(self, id=id, description=description,
                                        max_number_of_parallel_tasks=max_number_of_parallel_tasks,
                                        commands_before=commands_before, commands_after=commands_after)
        sql._SQLCommand.__init__(self, sql_statement, file_name, replace)

        self.parameter_function = parameter_function
        self.parameter_placeholders = parameter_placeholders
        self._db_alias = db_alias
        self.timezone = timezone
        self.echo_queries = echo_queries

    @property
    def db_alias(self):
        return self._db_alias or config.default_db_alias()

    def add_parallel_tasks(self, sub_pipeline: 'pipelines.Pipeline') -> None:
        parameters = self.parameter_function()

        if not isinstance(parameters, list) or not all(isinstance(item, tuple) for item in parameters):
            raise ValueError(f'parameter function should return a list of tuples, got "{repr(parameters)}"')

        for parameter_tuple in parameters:
            id = '-'.join([re.sub('[^0-9a-z\-_]+', '', str(x).lower().replace('-', '_')) for x in parameter_tuple])
            replace = self.replace.copy()
            for placeholder, param in zip(self.parameter_placeholders, parameter_tuple):
                replace[placeholder] = param

            sub_pipeline.add(pipelines.Task(
                id=id, description=f'Execute SQL for parameters {repr(parameter_tuple)}',
                commands=[
                    sql.ExecuteSQL(sql_file_name=self.sql_file_name, db_alias=self.db_alias,
                                   echo_queries=self.echo_queries, timezone=self.timezone, replace=replace)
                    if self.sql_file_name else
                    sql.ExecuteSQL(sql_statement=self.sql_statement, db_alias=self.db_alias,
                                   echo_queries=self.echo_queries, timezone=self.timezone, replace=replace)]))

    def html_doc_items(self) -> [(str, str)]:
        return [('db', _.tt[self.db_alias])] \
               + sql._SQLCommand.html_doc_items(self, self.db_alias) \
               + [('parameter function', html.highlight_syntax(inspect.getsource(self.parameter_function), 'python')),
                  ('parameter placeholders', _.tt[repr(self.parameter_placeholders)]),
                  ('echo queries', _.tt[str(self.echo_queries)]),
                  ('timezone', _.tt[self.timezone or ''])]



"""Functions for visualizing the last runs of a pipeline node"""

import datetime
import json

import flask
import psycopg2.extensions

import mara_db.postgresql
from mara_page import bootstrap, html, acl, _
from . import views
from .. import pipelines


def card(node: pipelines.Node) -> str:
    """A card that shows the system stats, the time line and output for the last runs or a node"""
    return bootstrap.card(
        id='last-runs-card',
        header_left=[
            'Last runs ',
            _.div(style='display:inline-block;margin-left:20px;')[html.asynchronous_content(
                flask.url_for('data_integration.last_runs_selector', path=node.url_path()))]],
        body=[html.spinner_js_function(),
              html.asynchronous_content(
                  url=flask.url_for('data_integration.system_stats', path=node.url_path(), run_id=None),
                  div_id='system-stats'),
              html.asynchronous_content(
                  url=flask.url_for('data_integration.timeline_chart', path=node.url_path(), run_id=None),
                  div_id='timeline-chart'),
              html.asynchronous_content(
                  url=flask.url_for('data_integration.run_output', path=node.url_path(), run_id=None, limit=True),
                  div_id='run-output')])


@views.blueprint.route('/<path:path>/last-runs-selector')
@views.blueprint.route('/last-runs-selector', defaults={'path': ''})
@acl.require_permission(views.acl_resource, do_abort=False)
def last_runs_selector(path: str):
    """
    Returns a html select element for selecting among the last runs of a node

    Args:
        path: The path of the node

    Returns:
        A `<select..><option ../><option ../></select>` element
    """
    from ..logging import node_cost

    node, __ = pipelines.find_node(path.split('/'))

    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
        cursor.execute(f'''
SELECT
  run_id,
  to_char(start_time, 'Mon DD HH24:MI') AS start_time,
  extract(EPOCH FROM (end_time - start_time)) AS duration,
  succeeded
FROM data_integration_node_run
WHERE node_path = {"%s"}
ORDER BY run_id DESC;''', (node.path(),))

        return str(
            _.select(id='last-runs-selector', class_='custom-select', style="border:none",
                     onchange=f"nodePage.switchRun(this.value, '{path}')")[
                [_.option(value=str(run_id))[
                     start_time, ' (',
                     f'{node_cost.format_duration(duration)}, {"succeeded" if succeeded else "failed"}'
                     if succeeded is not None else 'unfinished',
                     ')']
                 for run_id, start_time, duration, succeeded in cursor.fetchall()]])


@views.blueprint.route('/<path:path>/run-output', defaults={'run_id': None, 'limit': False})
@views.blueprint.route('/<path:path>/run-output/<int:run_id>', defaults={'limit': False})
@views.blueprint.route('/run-output', defaults={'path': '', 'run_id': None, 'limit': False})
@views.blueprint.route('/run-output/<int:run_id>', defaults={'path': '', 'limit': False})
@views.blueprint.route('/<path:path>/run-output-limited', defaults={'run_id': None, 'limit': True})
@views.blueprint.route('/<path:path>/run-output-limited/<int:run_id>', defaults={'limit': True})
@views.blueprint.route('/run-output-limited', defaults={'path': '', 'run_id': None, 'limit': True})
@views.blueprint.route('/run-output-limited/<int:run_id>', defaults={'path': '', 'limit': True})
@acl.require_permission(views.acl_resource, do_abort=False)
def run_output(path: str, run_id: int, limit: bool):
    """
    Returns the output of a node and its children as html

    Args:
        path: The path of the node
        run_id: The id of the run to return. If None, then the latest run is returned

    Returns:
        A <div class="run-output">..</div> element
    """
    node, __ = pipelines.find_node(path.split('/'))

    run_id = run_id or _latest_run_id(node.path())

    if not run_id:
        return ''

    line_limit = 1000
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
        cursor.execute(f'''
SELECT node_path, message, format, is_error
FROM data_integration_node_run
  JOIN data_integration_node_output USING (node_run_id)
WHERE node_path [1:{"%s"}] = %s 
      AND run_id = %s
ORDER BY timestamp 
''' + ('LIMIT ' + str(line_limit + 1) if limit else ''), (len(node.path()), node.path(), run_id))

        rows = cursor.fetchall()
        return str(_.script[f"""
nodePage.showOutput({json.dumps(rows[:line_limit] if limit else rows)},
               "{path}", 
               {'true' if len(rows) == line_limit + 1 else 'false'});
"""])


@views.blueprint.route('/<path:path>/system-stats', defaults={'run_id': None})
@views.blueprint.route('/<path:path>/system-stats/<int:run_id>')
@views.blueprint.route('/system-stats', defaults={'path': '', 'run_id': None})
@views.blueprint.route('/system-stats/<int:run_id>', defaults={'path': ''})
@acl.require_permission(views.acl_resource, do_abort=False)
def system_stats(path: str, run_id: int):
    node, __ = pipelines.find_node(path.split('/'))

    run_id = run_id or _latest_run_id(node.path())

    if not run_id:
        return ''

    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
        cursor.execute(f'''
SELECT data_integration_system_statistics.*
FROM data_integration_system_statistics
  JOIN data_integration_node_run ON timestamp BETWEEN start_time AND end_time
WHERE run_id = {"%s"} AND node_path = {"%s"};''', (run_id, node.path()))

        data = [[row[0].isoformat()] + list(row[1:]) for row in cursor.fetchall()]
        if len(data) >= 15:
            return str(_.div(id='system-stats-chart', class_='google-chart')[' ']) \
                   + str(_.script[f'nodePage.showSystemStats({json.dumps(data)});'])
        else:
            return ''


@views.blueprint.route('/<path:path>/timeline-chart', defaults={'run_id': None})
@views.blueprint.route('/<path:path>/timeline-chart/<int:run_id>')
@views.blueprint.route('/timeline-chart', defaults={'path': '', 'run_id': None})
@views.blueprint.route('/timeline-chart/<int:run_id>', defaults={'path': ''})
@acl.require_permission(views.acl_resource, do_abort=False)
def timeline_chart(path: str, run_id: int):
    node, __ = pipelines.find_node(path.split('/'))

    run_id = run_id or _latest_run_id(node.path())

    if not run_id:
        return ''

    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
        cursor.execute(f'''
SELECT node_path, start_time, end_time, max(end_time) over () AS max_end_time, succeeded, is_pipeline 
FROM data_integration_node_run
WHERE node_path [1 :{'%(level)s'}] = {'%(node_path)s'}
      AND array_length(node_path, 1) > {'%(level)s'}  
      AND run_id = {'%(run_id)s'};''', {'level': len(node.path()), 'node_path': node.path(), 'run_id': run_id})

        nodes = [{'label': ' / '.join(node_path[len(node.path()):]),
                  'status': {None: 'unfinished', True: 'succeeded', False: 'failed'}[succeeded],
                  'type': 'pipeline' if is_pipeline else 'task',
                  'url': flask.url_for('data_integration.node_page', path='/'.join(node_path)),
                  'start': start_time.isoformat(),
                  'end': (end_time or ((max_end_time or start_time) + datetime.timedelta(seconds=1))).isoformat()}
                 for node_path, start_time, end_time, max_end_time, succeeded, is_pipeline
                 in cursor.fetchall()]

        if nodes:
            return str(_.script[f"drawTimelineChart('timeline-chart', {json.dumps(nodes)})"])
        else:
            return ''


def _latest_run_id(node_path: [str]):
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:  # type: psycopg2.extensions.cursor
        cursor.execute('SELECT max(run_id) FROM data_integration_node_run WHERE node_path=%s', (node_path,))
        return cursor.fetchone()[0]



"""Data integration web UI"""

import functools

import flask

from mara_page import acl, navigation, _
from .. import config, pipelines

blueprint = flask.Blueprint('data_integration', __name__, url_prefix='/data-integration', static_folder='static')

acl_resource = acl.AclResource(name='Data Integration')


@functools.singledispatch
def navigation_icon(node: pipelines.Node):
    """Returns the navigation entry icon for a pipeline node"""
    return 'question'


navigation_icon.register(pipelines.Task)(lambda _: 'play')
navigation_icon.register(pipelines.ParallelTask)(lambda _: 'forward')
navigation_icon.register(pipelines.Pipeline)(lambda _: 'random')


def navigation_entry():
    """Creates a navigation entry that contains links to all data pipelines and their nodes"""

    def node_entry(node: pipelines.Node) -> navigation.NavigationEntry:
        return navigation.NavigationEntry(
            label=node.id, description=node.description, icon=navigation_icon(node),
            uri_fn=functools.partial(lambda n: node_url(n), node),
            children=([navigation.NavigationEntry(label='Overview', icon='list',
                                                  uri_fn=functools.partial(lambda: node_url(node)))]
                      + [node_entry(node) for node in node.nodes.values()]
                      if isinstance(node, pipelines.Pipeline) else None))

    entry = node_entry(config.root_pipeline())
    entry.label = 'Data Integration'
    entry.icon = 'wrench'
    entry.description = 'Pipelines for loading and transforming data'
    return entry


def node_url(node: pipelines.Node) -> str:
    """The url of the page that documents a node"""
    return flask.url_for('data_integration.node_page', path=node.url_path())


def format_labels(node: pipelines.Node):
    """Html markup that comma-separates labels of a node"""
    return ', '.join([str(_.span[label, ': ', _.tt(style='white-space:nowrap')[repr(value)]])
                      for label, value in node.labels.items()]) or ''



"""Generation of system statistics events (cpu, io, net, ram)"""

import datetime
import multiprocessing
import time

from .. import config
from ..logging import pipeline_events


class SystemStatistics(pipeline_events.Event):
    def __init__(self, timestamp: datetime.datetime, *, disc_read: float = None, disc_write: float = None,
                 net_recv: float = None, net_sent: float = None,
                 cpu_usage: float = None, mem_usage: float = None, swap_usage: float = None,
                 iowait: float = None) -> None:
        """
        Statistics about the system which runs the pipeline

        Individual statistics can be None

        Args:
            timestamp: The time when the statistics where gathered
            disc_read: read IO for discs in MB/s (summed)
            disc_write: write IO for discs in MB/s (summed)
            net_recv: read IO on all network adapters in MB/s (summed)
            net_sent: write IO on all network adapters in MB/s (summed)
            cpu_usage: cpu load on all cores in percent (summed)
            mem_usage: RAM used in percent of total ram
            swap_usage: swap used in percent of total swap
            iowait: How much time the CPU spends waiting for IO
        """
        super().__init__()
        self.timestamp = timestamp
        self.disc_read = disc_read
        self.disc_write = disc_write
        self.net_recv = net_recv
        self.net_sent = net_sent
        self.cpu_usage = cpu_usage
        self.mem_usage = mem_usage
        self.swap_usage = swap_usage
        self.iowait = iowait


def generate_system_statistics(event_queue: multiprocessing.Queue) -> None:
    """
    Generates one SystemStatistics event per configurable period and puts them in to a queue

    Ideas from
    http://off-the-stack.moorman.nu/2013-09-28-gather-metrics-using-psutil.html
    https://github.com/giampaolo/psutil/tree/master/scripts

    :param event_queue: The queue to write the events to
    """
    import psutil

    def cpu_usage():
        cpu_times = psutil.cpu_times_percent()
        return cpu_times.user + cpu_times.system

    def mem_usage():
        mem = psutil.virtual_memory()
        return 100.0 * mem.used / mem.total

    def swap_usage():
        swap = psutil.swap_memory()
        return 100.0 * swap.used / swap.total if swap.total > 0 else None

    # immediately send event for current cpu, mem and swap usage
    event_queue.put(SystemStatistics(
        datetime.datetime.now(), cpu_usage=cpu_usage(), mem_usage=mem_usage(), swap_usage=swap_usage()))
    period = config.system_statistics_collection_period()

    n = 0

    # some counters on WSL1 return None because psutil thinks it's linux,
    # but the linux kernel API is not implemented and fails
    # This lets it always return 0 for all attributes on that counter and lets at least CPU show up
    class _zero():
        def __getattr__(self, item): return 0
    zero = _zero()

    # capture current disc and net state for later diff
    discs_last = psutil.disk_io_counters() or zero
    nets_last = psutil.net_io_counters() or zero
    mb = 1024 * 1024
    time.sleep(period)
    while True:
        discs_cur = psutil.disk_io_counters() or zero
        nets_cur = psutil.net_io_counters() or zero
        event_queue.put(SystemStatistics(
            datetime.datetime.now(),
            disc_read=(discs_cur.read_bytes - discs_last.read_bytes) / mb / period,
            disc_write=(discs_cur.write_bytes - discs_last.write_bytes) / mb / period,
            net_recv=(nets_cur.bytes_recv - nets_last.bytes_recv) / mb / period,
            net_sent=(nets_cur.bytes_sent - nets_last.bytes_sent) / mb / period,
            cpu_usage=cpu_usage(), mem_usage=mem_usage(), swap_usage=swap_usage()))
        nets_last = nets_cur
        discs_last = discs_cur

        # double period every 100 measurements in order to avoid sending too many requests to frontend
        n += 1
        if n % 100 == 0:
            period *= 2

        time.sleep(period)



"""
Execution of data pipelines.
Uses forking (multiprocessing processes) for parallelism and message queues for inter-process communication.
"""

import datetime
import functools
import multiprocessing
import os
import signal
import time
import traceback
from multiprocessing import queues

from . import pipelines, config
from .logging import logger, pipeline_events, system_statistics, run_log, node_cost
from . import events


def run_pipeline(pipeline: pipelines.Pipeline, nodes: {pipelines.Node} = None,
                 with_upstreams: bool = False,
                 interactively_started: bool = False
                 ) -> [events.Event]:
    """
    Runs a pipeline in a forked sub process. Acts as a generator that yields events from the sub process.

    Using forking has two advantages:
    1. The pipeline is also forked and thus can be modified without affecting the original pipeline.
    2. It's possible to hand over control to the parent process while the pipeline is running, for example
       for sending output to a browser.

    Args:
        pipeline: The pipeline to run
        nodes: A list of pipeline children that should run
        with_upstreams: When true and `nodes` are provided, then all upstreams of `nodes` in `pipeline` are also run
    Yields:
        Events emitted during pipeline execution
    """

    # use forking for starting child processes to avoid cleanup functions and leakage and pickle problems
    #
    # On newer macs you need to set
    #   OBJC_DISABLE_INITIALIZE_FORK_SAFETY=YES
    # env variable *before* starting python/flask otherwise you will get core dumps when any forked process calls
    # into certain native code (e.g. requests)! Note that this is done automatically if you create your virtual env
    # via the scripts from mara-app >= 2.1.1
    #
    multiprocessing_context = multiprocessing.get_context('fork')

    # A queue for receiving events from forked sub processes
    event_queue = multiprocessing_context.Queue()

    # The function that is run in a sub process
    def run():

        # collect system stats in a separate Process
        statistics_process = multiprocessing_context.Process(
            target=lambda: system_statistics.generate_system_statistics(event_queue), name='system_statistics')
        statistics_process.start()

        try:
            # capture output of print statements and other unplanned output
            logger.redirect_output(event_queue, pipeline.path())

            # all nodes that have not run yet, ordered by priority
            node_queue: [pipelines.Node] = []

            # data needed for computing cost
            node_durations_and_run_times = node_cost.node_durations_and_run_times(pipeline)

            # Putting nodes into the node queue
            def queue(nodes: [pipelines.Node]):
                for node in nodes:
                    node_cost.compute_cost(node, node_durations_and_run_times)
                    node_queue.append(node)
                node_queue.sort(key=lambda node: node.cost, reverse=True)

            if nodes:  # only run a set of child nodes
                def with_all_upstreams(nodes: {pipelines.Node}):
                    """recursively find all upstreams of a list of nodes"""
                    return functools.reduce(set.union, [with_all_upstreams(node.upstreams) for node in nodes], nodes)

                # when requested, include all upstreams of nodes, otherwise just use provided nodes
                nodes_to_run = with_all_upstreams(set(nodes)) if with_upstreams else set(nodes)

                # remove everything from pipeline that should not be run
                # (that's makes updating dependencies between nodes easier)
                for node in set(pipeline.nodes.values()) - nodes_to_run:
                    pipeline.remove(node)

                # queue remaining nodes
                queue(list((pipeline.nodes).values()))

            else:
                # remove dependencies to siblings
                pipeline.upstreams = set()
                pipeline.downstreams = set()
                # queue whole pipeline
                queue([pipeline])

            # book keeping
            run_start_time = datetime.datetime.now()
            # all nodes that already ran or that won't be run anymore
            processed_nodes: {pipelines.Node} = set()
            # running pipelines with start times and number of running children
            running_pipelines: {pipelines.Pipeline: [datetime.datetime, int]} = {}
            failed_pipelines: {pipelines.Pipeline} = set()  # pipelines with failed tasks
            running_task_processes: {pipelines.Task: TaskProcess} = {}

            def dequeue() -> pipelines.Node:
                """
                Finds the next task in the queue
                - without upstreams or where all upstreams have been run already
                - where the pipeline specific maximum number of parallel tasks per pipeline is not reached
                """
                for node in node_queue:  # type: pipelines.Node
                    if ((not node.upstreams or len(node.upstreams & processed_nodes) == len(node.upstreams))
                        and (not isinstance(node.parent, pipelines.Pipeline)
                             or (not node.parent.max_number_of_parallel_tasks)
                             or (not node.parent in running_pipelines)
                             or (running_pipelines[node.parent][1] < node.parent.max_number_of_parallel_tasks))):
                        node_queue.remove(node)
                        if node.parent in failed_pipelines and not node.parent.force_run_all_children:
                            # if the parent pipeline failed (and no overwrite), don't launch new nodes
                            processed_nodes.add(node)
                        else:
                            return node

            def track_finished_pipelines():
                """when all nodes of a pipeline have been processed, then emit events"""
                for running_pipeline, (start_time, running_children) \
                    in dict(running_pipelines).items():  # type: pipelines.Pipeline
                    if len(set(running_pipeline.nodes.values()) & processed_nodes) == len(running_pipeline.nodes):
                        succeeded = running_pipeline not in failed_pipelines
                        event_queue.put(pipeline_events.Output(
                            node_path=running_pipeline.path(), format=logger.Format.ITALICS, is_error=not succeeded,
                            message=f'{"succeeded" if succeeded else "failed"}, {logger.format_time_difference(run_start_time, datetime.datetime.now())}'))
                        event_queue.put(pipeline_events.NodeFinished(
                            node_path=running_pipeline.path(), start_time=start_time,
                            end_time=datetime.datetime.now(), is_pipeline=True, succeeded=succeeded))
                        del running_pipelines[running_pipeline]
                        processed_nodes.add(running_pipeline)

            # announce run start
            event_queue.put(pipeline_events.RunStarted(node_path=pipeline.path(),
                                                       start_time=run_start_time,
                                                       pid=os.getpid(),
                                                       interactively_started=interactively_started,
                                                       node_ids=[node.id for node in (nodes or [])],
                                                       is_root_pipeline=(pipeline.parent is None))
                            )

            # run as long
            # - as task processes are still running
            # - as there is still stuff in the node queue
            while running_task_processes or node_queue:
                # don't do anything if the maximum number of parallel tasks is currently running
                if len(running_task_processes) < config.max_number_of_parallel_tasks():

                    next_node = dequeue()  # get the next runnable node from the queue

                    if next_node:
                        if isinstance(next_node, pipelines.Pipeline):
                            # connect pipeline nodes without upstreams to upstreams of pipeline
                            for upstream in next_node.upstreams:
                                for pipeline_node in next_node.nodes.values():
                                    if not pipeline_node.upstreams:
                                        next_node.add_dependency(upstream, pipeline_node)

                            # connect pipeline nodes without downstreams to downstream of pipeline
                            for downstream in next_node.downstreams:
                                for pipeline_node in next_node.nodes.values():
                                    if not pipeline_node.downstreams:
                                        next_node.add_dependency(pipeline_node, downstream)

                            # get cost information for children
                            node_durations_and_run_times.update(node_cost.node_durations_and_run_times(next_node))

                            # queue all child nodes
                            queue(list(next_node.nodes.values()))

                            # book keeping and event emission
                            pipeline_start_time = datetime.datetime.now()
                            running_pipelines[next_node] = [pipeline_start_time, 0]
                            event_queue.put(pipeline_events.NodeStarted(next_node.path(), pipeline_start_time, True))
                            event_queue.put(pipeline_events.Output(
                                node_path=next_node.path(), format=logger.Format.ITALICS,
                                message='★ ' + node_cost.format_duration(
                                    node_durations_and_run_times.get(tuple(next_node.path()), [0, 0])[0])))

                        elif isinstance(next_node, pipelines.ParallelTask):
                            # create sub tasks and queue them
                            task_start_time = datetime.datetime.now()
                            try:
                                logger.redirect_output(event_queue, next_node.path())
                                logger.log('☆ Launching tasks', format=logger.Format.ITALICS)
                                sub_pipeline = next_node.launch()
                                next_node.parent.replace(next_node, sub_pipeline)
                                queue([sub_pipeline])

                            except Exception as e:
                                event_queue.put(pipeline_events.NodeStarted(
                                    node_path=next_node.path(), start_time=task_start_time, is_pipeline=True))
                                logger.log(message=f'Could not launch parallel tasks', format=logger.Format.ITALICS,
                                           is_error=True)
                                logger.log(message=traceback.format_exc(),
                                           format=pipeline_events.Output.Format.VERBATIM, is_error=True)
                                event_queue.put(pipeline_events.NodeFinished(
                                    node_path=next_node.path(), start_time=task_start_time,
                                    end_time=datetime.datetime.now(), is_pipeline=True, succeeded=False))

                                failed_pipelines.add(next_node.parent)
                                processed_nodes.add(next_node)
                            finally:
                                logger.redirect_output(event_queue, pipeline.path())

                        else:
                            # run a task in a subprocess
                            if next_node.parent in running_pipelines:
                                running_pipelines[next_node.parent][1] += 1
                            event_queue.put(
                                pipeline_events.NodeStarted(next_node.path(), datetime.datetime.now(), False))
                            event_queue.put(pipeline_events.Output(
                                node_path=next_node.path(), format=logger.Format.ITALICS,
                                message='★ ' + node_cost.format_duration(
                                    node_durations_and_run_times.get(tuple(next_node.path()), [0, 0])[0])))

                            status_queue = multiprocessing_context.Queue()
                            process = TaskProcess(next_node, event_queue, status_queue)
                            process.start()
                            running_task_processes[next_node] = process

                # check whether some of the running processes finished
                for task_process in list(running_task_processes.values()):  # type: TaskProcess
                    if task_process.is_alive():
                        pass
                    else:
                        del running_task_processes[task_process.task]
                        if task_process.task.parent in running_pipelines:
                            running_pipelines[task_process.task.parent][1] -= 1

                        processed_nodes.add(task_process.task)

                        succeeded = not (task_process.status_queue.get() == False or task_process.exitcode != 0)
                        if not succeeded and not task_process.task.parent.ignore_errors:
                            for parent in task_process.task.parents()[:-1]:
                                failed_pipelines.add(parent)

                        end_time = datetime.datetime.now()
                        event_queue.put(
                            pipeline_events.Output(task_process.task.path(),
                                                   ('succeeded' if succeeded else 'failed') + ',  '
                                                   + logger.format_time_difference(task_process.start_time, end_time),
                                                   format=logger.Format.ITALICS, is_error=not succeeded))
                        event_queue.put(pipeline_events.NodeFinished(task_process.task.path(), task_process.start_time,
                                                                     end_time, False, succeeded))

                # check if some pipelines finished
                track_finished_pipelines()

                # don't busy-wait
                time.sleep(0.001)

        except:
            event_queue.put(pipeline_events.Output(node_path=pipeline.path(), message=traceback.format_exc(),
                                                   format=logger.Format.ITALICS, is_error=True))

        # run again because `dequeue` might have moved more nodes to `finished_nodes`
        track_finished_pipelines()

        # kill the stats process (joining or terminating does not work in gunicorn)
        os.kill(statistics_process.pid, signal.SIGKILL)
        statistics_process.join()

        # run finished
        event_queue.put(pipeline_events.RunFinished(node_path=pipeline.path(), end_time=datetime.datetime.now(),
                                                    succeeded=not failed_pipelines,
                                                    interactively_started=interactively_started))

    # fork the process and run `run`
    run_process = multiprocessing_context.Process(target=run, name='pipeline-' + '-'.join(pipeline.path()))
    run_process.start()

    runlogger = run_log.RunLogger()

    # process messages from forked child processes
    while True:
        try:
            while not event_queue.empty():
                event = event_queue.get(False)
                runlogger.handle_event(event)
                events.notify_configured_event_handlers(event)
                yield event
        except queues.Empty:
            pass
        except GeneratorExit:
            # This happens e.g. if the browser window is closed or we reload the page in the middle of a run
            # As we still have open runs, we need to close them as failed.
            run_log.close_open_run_after_error(runlogger.run_id)
            # Catching GeneratorExit needs to end in a return!
            return
        except:
            output_event = pipeline_events.Output(node_path=pipeline.path(), message=traceback.format_exc(),
                                                  format=logger.Format.ITALICS, is_error=True)
            runlogger.handle_event(output_event)
            events.notify_configured_event_handlers(output_event)
            yield output_event
            run_log.close_open_run_after_error(runlogger.run_id)
            return
        if not run_process.is_alive():
            break
        time.sleep(0.001)


class TaskProcess(multiprocessing.Process):
    def __init__(self, task: pipelines.Task, event_queue: multiprocessing.Queue, status_queue: multiprocessing.Queue):
        """
        Runs a task in a separate sub process.

        Args:
            task: The task to run
            event_queue: The query for writing events to
            status_queue: A queue for reporting whether the task succeeded
        """
        super().__init__(name='task-' + '-'.join(task.path()))
        self.task = task
        self.event_queue = event_queue
        self.status_queue = status_queue
        self.start_time = datetime.datetime.now()

    def run(self):
        # redirect stdout and stderr to queue
        logger.redirect_output(self.event_queue, self.task.path())

        succeeded = True
        attempt = 0
        try:
            while True:
                if not self.task.run():
                    if attempt < self.task.max_retries:
                        attempt += 1
                        delay = pow(2, attempt + 2)
                        logger.log(message=f'Retry {attempt}/{self.task.max_retries} in {delay} seconds',
                                   is_error=True, format=logger.Format.ITALICS)
                        time.sleep(delay)
                    else:
                        succeeded = False
                        break
                else:
                    break
        except Exception as e:
            logger.log(message=traceback.format_exc(), format=logger.Format.VERBATIM, is_error=True)
            succeeded = False

        self.status_queue.put(succeeded)



"""Events that are emitted during pipeline execution"""

import datetime
import json

import enum
import typing as t

from ..events import Event


class PipelineEvent(Event):
    def __init__(self, node_path: [str]) -> None:
        """
        Base class for events that are emitted during a pipeline run.

        Args:
            node_path: The path of the current node in the data pipeline that is run

        """
        super().__init__()
        self.node_path = node_path

    def to_json(self):
        return json.dumps({field: value.isoformat() if isinstance(value, datetime.datetime) else value
                           for field, value in self.__dict__.items()})


class RunStarted(PipelineEvent):
    def __init__(self, node_path: [str],
                 start_time: datetime.datetime,
                 pid: int,
                 is_root_pipeline: bool = False,
                 node_ids: t.Optional[t.List[str]] = None,
                 interactively_started: bool = False) -> None:
        """
        A pipeline run started
        Args:
            node_path: The path of the pipeline that was run
            start_time: The time when the run started
            pid: The process id of the process that runs the pipeline
            node_ids: list of node.ids which should be run
            is_root_pipeline: whether this pipeline run runs the root pipeline
            interactively_started: whether or not the run was started interactively
        """
        super().__init__([])
        self.node_path = node_path
        self.start_time = start_time
        self.pid = pid
        self.interactively_started = interactively_started
        self.is_root_pipeline = is_root_pipeline
        self.node_ids = node_ids or []

        self.user: str = get_user_display_name(interactively_started)


class RunFinished(PipelineEvent):
    def __init__(self, node_path: [str],
                 end_time: datetime.datetime,
                 succeeded: bool,
                 interactively_started: bool = False) -> None:
        """
        A pipeline run finished
        Args:
            node_path: The path of the pipeline that was run
            end_time: The time when the run finished
            succeeded: Whether the run succeeded
            interactively_started: whether or not the run was started interactively
        """
        super().__init__([])
        self.node_path = node_path
        self.end_time = end_time
        self.succeeded = succeeded
        self.interactively_started = interactively_started


class NodeStarted(PipelineEvent):
    def __init__(self, node_path: [str], start_time: datetime.datetime, is_pipeline: bool) -> None:
        """
        A task run started.
        Args:
            node_path: The path of the current node in the data pipeline that is run
            start_time: The time when the task started
            is_pipeline: Whether the node is a pipeline
        """
        super().__init__(node_path)
        self.start_time = start_time
        self.is_pipeline = is_pipeline


class NodeFinished(PipelineEvent):
    def __init__(self, node_path: [str], start_time: datetime.datetime, end_time: datetime.datetime,
                 is_pipeline: bool, succeeded: bool) -> None:
        """
        A run of a task or pipeline finished.
        Args:
            node_path: The path of the current node in the data pipeline that is run
            start_time: The time when the task started
            end_time: The time when the task finished
            is_pipeline: Whether the node is a pipeline
            succeeded: Whether the task succeeded
        """
        super().__init__(node_path)
        self.start_time = start_time
        self.end_time = end_time
        self.is_pipeline = is_pipeline
        self.succeeded = succeeded


class Output(PipelineEvent):
    class Format(enum.EnumMeta):
        """Formats for displaying log messages"""
        STANDARD = 'standard'
        VERBATIM = 'verbatim'
        ITALICS = 'italics'

    def __init__(self, node_path: [str], message: str,
                 format: Format = Format.STANDARD, is_error: bool = False) -> None:
        """
        Some text output occurred.
        Args:
            node_path: The path of the current node in the data pipeline that is run
            message: The message to display
            format: How to format the message
            is_error: Whether the message is considered an error message
        """
        super().__init__(node_path)
        self.message = message
        self.format = format
        self.is_error = is_error
        self.timestamp = datetime.datetime.now()


def get_user_display_name(interactively_started:bool) -> t.Optional[str]:
    """Gets the display name for the user which started a run

    Defaults to MARA_RUN_USER_DISPLAY_NAME and falls back to the current OS-level name
    if the run was started interactively, else None.

    :param interactively_started: True if the run was triggered interactively

    Patch if you have more sophisticated needs.
    """
    import os
    if 'MARA_RUN_USER_DISPLAY_NAME' in os.environ:
        return os.environ.get('MARA_RUN_USER_DISPLAY_NAME')
    if not interactively_started:
        return None
    return os.environ.get('SUDO_USER') or os.environ.get('USER') or os.getlogin()



"""Commands for running bash scripts"""

from typing import Union, Callable

from mara_page import html
from .. import pipelines


class RunBash(pipelines.Command):
    def __init__(self, command: Union[str, Callable]) -> None:
        """
        Runs a command in a bash shell

        Args:
            command: The command to run
        """
        super().__init__()
        self._command = command

    @property
    def command(self):
        return (self._command() if callable(self._command) else self._command).strip()

    def shell_command(self):
        return self.command

    def html_doc_items(self) -> [(str, str)]:
        return [
            ('command', html.highlight_syntax(self.shell_command(), 'bash'))
        ]



"""Resetting incremental copy status"""

import mara_db.config
import mara_db.postgresql


def reset_incremental_processing(node_path: [str]):
    """
    Recursively resets all incremental processing status information that is stored in the mara db
    Args:
        node_path: The path of the node to reset

    """
    with mara_db.postgresql.postgres_cursor_context('mara') as cursor:
        cursor.execute(f'''
SELECT *
FROM
  (SELECT node_path, 'processed files', count(*)
   FROM data_integration_processed_file
   GROUP BY node_path

   UNION

   SELECT node_path, 'file dependencies', count(*)
   FROM data_integration_file_dependency
   GROUP BY node_path

   UNION

   SELECT node_path, 'incremental copy statuses', count(*)
   FROM data_integration_incremental_copy_status
   GROUP BY node_path

  ) t
WHERE node_path [1:{'%s'}] = {'%s'}

ORDER BY 1, 2''', (len(node_path), node_path))

        for path, type, n in cursor.fetchall():
            print(f'{"/".join(path)}: {n} {type}')

        for table in ['data_integration_processed_file', 'data_integration_file_dependency',
                      'data_integration_incremental_copy_status']:
            cursor.execute(f'''DELETE FROM {table} WHERE node_path [1:{'%s'}] = {'%s'}''', (len(node_path), node_path))



"""Commands for working with sql databases"""

import functools
import json
import pathlib
import shlex
from typing import Callable, Union

import mara_db.dbs
import mara_db.shell
from mara_page import _, html
from .. import config, shell, pipelines
from ..incremental_processing import file_dependencies
from ..incremental_processing import incremental_copy_status
from ..logging import logger


class _SQLCommand(pipelines.Command):
    def __init__(self, sql_statement: Union[Callable, str] = None, sql_file_name: str = None,
                 replace: {str: str} = None) -> None:
        """
        Something that runs a sql query (either a query string or a query file).

        Args:
            sql_statement: The statement to run as a string
            sql_file_name: The name of the file to run (relative to the directory of the parent pipeline)
            replace: A set of replacements to perform against the sql query `{'replace`: 'with', ..}`
        """
        if (not (sql_statement or sql_file_name)) or (sql_statement and sql_file_name):
            raise ValueError('Please provide either sql_statement or sql_file_name (but not both)')

        super().__init__()
        self.sql_file_name = sql_file_name
        self._sql_statement = sql_statement
        self.replace = replace or {}

    @property
    def sql_statement(self):
        return self._sql_statement() if callable(self._sql_statement) else self._sql_statement

    def sql_file_path(self) -> pathlib.Path:
        # Get the first pipeline in the tree (don't reach root)
        pipeline_candidate = self
        while not isinstance(pipeline_candidate, pipelines.Pipeline):
            pipeline_candidate = pipeline_candidate.parent
        return pipeline_candidate.base_path() / self.sql_file_name

    def shell_command(self):
        if self.sql_file_name:
            command = 'cat ' + shlex.quote(str(self.sql_file_path().absolute())) + ' \\\n'
        else:
            command = f'echo {shlex.quote(self.sql_statement)} \\\n'

        if self.replace:
            command += '  | ' + shell.sed_command(_expand_pattern_substitution(self.replace)) + ' \\\n'
        return command

    def html_doc_items(self, db_alias: str):
        sql = self.sql_statement or \
              (self.sql_file_path().read_text().strip('\n') if self.sql_file_path().exists() else '-- file not found')
        doc = []
        if self.sql_statement:
            doc.append(
                ('sql statement', html.highlight_syntax(self.sql_statement, _sql_syntax_higlighting_lexter(db_alias))))
        else:
            doc.append(('sql file name', _.i[self.sql_file_name]))
            doc.append((_.i['sql file content'], html.highlight_syntax(sql, _sql_syntax_higlighting_lexter(db_alias))))

        doc.append(('replace', html.highlight_syntax(
            json.dumps(_expand_pattern_substitution(self.replace), indent=2), 'json')))

        if self.replace:
            for k, v in _expand_pattern_substitution(self.replace).items():
                sql = sql.replace(k, v)
            doc.append((_.i['substituted sql'], html.highlight_syntax(sql, _sql_syntax_higlighting_lexter(db_alias))))

        return doc


class ExecuteSQL(_SQLCommand):
    def __init__(self, sql_statement: str = None, sql_file_name: Union[str, Callable] = None,
                 replace: {str: str} = None, file_dependencies=None, db_alias: str = None,
                 echo_queries: bool = True, timezone: str = None) -> None:
        """
        Runs an sql file or statement in a database

        Args:
            sql_statement: The statement to run as a string
            sql_file_name: The name of the file to run (relative to the directory of the parent pipeline)
            replace: A set of replacements to perform against the sql query `{'replace`: 'with', ..}`
        """
        _SQLCommand.__init__(self, sql_statement, sql_file_name, replace)

        self._db_alias = db_alias
        self.timezone = timezone
        self.file_dependencies = file_dependencies or []
        self.echo_queries = echo_queries

    @property
    def db_alias(self):
        return self._db_alias or config.default_db_alias()

    def run(self) -> bool:
        if self.sql_file_name:
            logger.log(self.sql_file_name, logger.Format.ITALICS)

        dependency_type = 'ExecuteSQL ' + (self.sql_file_name or self.sql_statement)

        if self.file_dependencies:
            assert (self.parent)
            pipeline_base_path = self.parent.parent.base_path()
            if not file_dependencies.is_modified(self.node_path(), dependency_type,
                                                 pipeline_base_path,
                                                 self.file_dependencies):
                logger.log('no changes')
                return True
            else:
                # delete any old hash to trigger a run in case the next run is switched back to the old hash
                # This prevents inconsistent state in case you do a deplyoment with bad SQL which fails and
                # then revert.
                # The hash would still be correct for the old state but the results of this file would
                # probably not be there (usually the first step is a DROP).
                file_dependencies.delete(self.node_path(), dependency_type)

        if not super().run():
            return False

        if self.file_dependencies:
            file_dependencies.update(self.node_path(), dependency_type, pipeline_base_path, self.file_dependencies)
        return True

    def shell_command(self):
        return _SQLCommand.shell_command(self) \
               + '  | ' + mara_db.shell.query_command(self.db_alias, self.timezone, self.echo_queries)

    def html_doc_items(self):
        return [('db', _.tt[self.db_alias]),
                ('file dependencies', [_.i[dependency, _.br] for dependency in self.file_dependencies])] \
               + _SQLCommand.html_doc_items(self, self.db_alias) \
               + [('echo queries', _.tt[str(self.echo_queries)]),
                  ('timezone', _.tt[self.timezone or '']),
                  (_.i['shell command'], html.highlight_syntax(self.shell_command(), 'bash'))]


class Copy(_SQLCommand):
    """Loads data from an external database"""

    def __init__(self, source_db_alias: str, target_table: str, target_db_alias: str = None,
                 sql_statement: str = None, sql_file_name: Union[Callable, str] = None, replace: {str: str} = None,
                 timezone: str = None, csv_format: bool = None, delimiter_char: str = None,
                 file_dependencies=None) -> None:
        _SQLCommand.__init__(self, sql_statement, sql_file_name, replace)
        self.source_db_alias = source_db_alias
        self.target_table = target_table
        self._target_db_alias = target_db_alias
        self.timezone = timezone
        self.csv_format = csv_format
        self.delimiter_char = delimiter_char
        self.file_dependencies = file_dependencies or []

    @property
    def target_db_alias(self):
        return self._target_db_alias or config.default_db_alias()

    def file_path(self) -> pathlib.Path:
        return self.parent.parent.base_path() / self.file_name

    def run(self) -> bool:
        if self.sql_file_name:
            logger.log(self.sql_file_name, logger.Format.ITALICS)

        dependency_type = 'Copy ' + (self.sql_file_name or self.sql_statement)

        if self.file_dependencies:
            assert (self.parent)
            pipeline_base_path = self.parent.parent.base_path()
            if not file_dependencies.is_modified(self.node_path(), dependency_type,
                                                 pipeline_base_path,
                                                 self.file_dependencies):
                logger.log('no changes')
                return True
            else:
                # delete any old hash to trigger a run in case the next run is switched back to the old hash
                # which in most cases would result in an newly created empty table but no load
                # (see also above in ExecuteSQL)
                file_dependencies.delete(self.node_path(), dependency_type)


        if not super().run():
            return False

        if self.file_dependencies:
            file_dependencies.update(self.node_path(), dependency_type, pipeline_base_path, self.file_dependencies)
        return True

    def shell_command(self):
        return _SQLCommand.shell_command(self) \
               + '  | ' + mara_db.shell.copy_command(self.source_db_alias, self.target_db_alias, self.target_table,
                                                     self.timezone, self.csv_format, self.delimiter_char)

    def html_doc_items(self) -> [(str, str)]:
        return [('source db', _.tt[self.source_db_alias])] \
               + _SQLCommand.html_doc_items(self, self.source_db_alias) \
               + [('target db', _.tt[self.target_db_alias]),
                  ('target table', _.tt[self.target_table]),
                  ('timezone', _.tt[self.timezone or '']),
                  ('csv format', _.tt[self.csv_format or '']),
                  ('delimiter char', _.tt[self.delimiter_char or '']),
                  (_.i['shell command'], html.highlight_syntax(self.shell_command(), 'bash'))]


class CopyIncrementally(_SQLCommand):
    def __init__(self, source_db_alias: str, source_table: str,
                 modification_comparison: str, comparison_value_placeholder: str,
                 target_table: str, primary_keys: [str],
                 sql_file_name: Union[str, Callable] = None, sql_statement: Union[str, Callable] = None,
                 target_db_alias: str = None, timezone: str = None, replace: {str: str} = None,
                 use_explicit_upsert: bool = False,
                 csv_format: bool = None, delimiter_char: str = None) -> None:
        """
        Incrementally loads data from one database into another.

        Requires the source table to have an monotonously increasing column or combination of columns that
        allow to identify "newer" columns (the modification comparison).

        After an initial full load, only those rows with a with a a higher modification comparison than the last
        comparison value are read.

        Args:
            source_db_alias: The database to load from
            source_table: The table to read from
            sql_statement: A query that is run to query the source database
            sql_file_name: The path of a file name that is run to query the source database
            replace: A set of replacements to perform against the sql query
            modification_comparison: SQL expression that evaluates to a comparable value
            comparison_value_placeholder: A placeholder that is replaced with the last comparision value in the sql query
            target_db_alias: The database to write to
            target_table: The table for loading data into
            primary_keys: A combination of primary key columns that are used for upserting into the target table
            timezone: How to interpret timestamps in the target db
            use_explicit_upsert: When True, uses an Update + Insert query combination. Otherwise ON CONFLICT DO UPDATE.
        """
        _SQLCommand.__init__(self, sql_statement, sql_file_name, replace)
        self.source_db_alias = source_db_alias
        self.source_table = source_table
        self.modification_comparison = modification_comparison
        self.comparison_value_placeholder = comparison_value_placeholder

        self._target_db_alias = target_db_alias
        self.target_table = target_table
        self.primary_keys = primary_keys
        self.timezone = timezone
        self.use_explicit_upsert = use_explicit_upsert
        self.csv_format = csv_format
        self.delimiter_char = delimiter_char

    @property
    def target_db_alias(self):
        return self._target_db_alias or config.default_db_alias()

    def run(self) -> bool:
        # retrieve the highest current value for the modification comparison (e.g.: the highest timestamp)
        # We intentionally use the command line here (rather than sqlalchemy) to avoid forcing people python drivers,
        # which can be hard for example in the case of SQL Server
        logger.log(f'get highest modification comparison value', format=logger.Format.ITALICS)
        max_value_query = f'SELECT max({self.modification_comparison}) AS maxval FROM {self.source_table}'
        logger.log(max_value_query, format=logger.Format.VERBATIM)
        result = shell.run_shell_command(f'echo {shlex.quote(max_value_query)} \\\n  | '
                                         + mara_db.shell.copy_to_stdout_command(self.source_db_alias))

        if not result:
            return False

        # be flexible with different output formats: remove the column header & remove whitespace & quotes
        max_modification_value = ''.join(result).replace('maxval', '').strip().strip('"')
        logger.log(repr(max_modification_value), format=logger.Format.VERBATIM)

        # check whether target table is empty
        target_table_is_empty = True

        target_table_empty_query = f'SELECT TRUE FROM {self.target_table} LIMIT 1'
        logger.log(f'check if target table is empty', format=logger.Format.ITALICS)
        logger.log(target_table_empty_query, format=logger.Format.VERBATIM)
        with mara_db.postgresql.postgres_cursor_context(self.target_db_alias) as cursor:
            cursor.execute(f'SELECT TRUE FROM {self.target_table} LIMIT 1')
            target_table_is_empty = not cursor.fetchone()
            logger.log(f"target table{'' if target_table_is_empty else ' not'} empty", format=logger.Format.ITALICS)

        # get last comparison value
        logger.log('get last comparison value', format=logger.Format.ITALICS)
        last_comparison_value = incremental_copy_status.get_last_comparison_value(
            self.node_path(), self.source_db_alias, self.source_table)
        logger.log(repr(last_comparison_value), format=logger.Format.VERBATIM)

        if target_table_is_empty or not last_comparison_value:
            # full load
            logger.log('full (non incremental) copy', logger.Format.ITALICS)
            if not target_table_is_empty:
                truncate_query = f'TRUNCATE TABLE {self.target_table}'
                logger.log(truncate_query, format=logger.Format.VERBATIM)
                with mara_db.postgresql.postgres_cursor_context(self.target_db_alias) as cursor:
                    cursor.execute(truncate_query)
            elif last_comparison_value:
                # table is empty but we have a last comparison value from earlier runs
                # If we would crash during load (with some data already in the table), the next run would
                # not trigger a full load and we would miss data. To prevent that, delete the old
                # comparison value (we will then set it only on success)
                logger.log('deleting old comparison value', logger.Format.ITALICS)
                incremental_copy_status.delete(self.node_path(), self.source_db_alias, self.source_table)

            # overwrite the comparison criteria to get everything
            replace = {self.comparison_value_placeholder: '(1=1)'}
            complete_copy_command = self._copy_command(self.target_table, replace)
            if not shell.run_shell_command(complete_copy_command):
                return False

        else:
            # incremental load. First create the table which will contain the delta
            logger.log('incremental copy, create upsert table', logger.Format.ITALICS)
            create_upsert_table_query = (f'DROP TABLE IF EXISTS {self.target_table}_upsert;\n'
                                         + f'CREATE TABLE {self.target_table}_upsert AS SELECT * from {self.target_table} WHERE FALSE')

            if not shell.run_shell_command(f'echo {shlex.quote(create_upsert_table_query)} \\\n  | '
                                           + mara_db.shell.query_command(self.target_db_alias, echo_queries=True)):
                return False

            # perform the actual copy replacing the placeholder
            # with the comparison value from the latest successful execution
            replace = {self.comparison_value_placeholder:
                           f'({self.modification_comparison} >= \'{last_comparison_value}\')'}
            if not shell.run_shell_command(self._copy_command(self.target_table + '_upsert', replace)):
                return False

            # now the upsert table has to be merged with the target one

            # retrieve the target table columns to build the SET clause of the upsert query
            with mara_db.postgresql.postgres_cursor_context(self.target_db_alias) as cursor:
                retrieve_column_query = f"SELECT attname FROM pg_attribute WHERE attrelid = '{self.target_table}'::REGCLASS AND attnum > 0;"
                logger.log(retrieve_column_query, format=logger.Format.VERBATIM)
                cursor.execute(retrieve_column_query)
                if self.use_explicit_upsert:
                    set_clause = ', '.join([f'"{col[0]}" = src."{col[0]}"' for col in cursor.fetchall()])
                    key_definition = ' AND '.join([f'dst."{k}" = src."{k}"' for k in self.primary_keys])
                else:
                    set_clause = ', '.join([f'"{col[0]}" = EXCLUDED."{col[0]}"' for col in cursor.fetchall()])
                    key_definition = ', '.join(['"' + primary_key + '"' for primary_key in self.primary_keys])

            if self.use_explicit_upsert:
                update_query = f"""
UPDATE {self.target_table} dst
SET {set_clause}
FROM {self.target_table}_upsert src
WHERE {key_definition}"""

                insert_query = f"""
INSERT INTO {self.target_table}
SELECT src.*
FROM {self.target_table}_upsert src
WHERE NOT EXISTS (SELECT 1 FROM {self.target_table} dst WHERE {key_definition})"""
                if not shell.run_shell_command(f'echo {shlex.quote(update_query)} \\\n  | '
                                               + mara_db.shell.query_command(self.target_db_alias, echo_queries=True)):
                    return False
                elif not shell.run_shell_command(f'echo {shlex.quote(insert_query)} \\\n  | '
                                                 + mara_db.shell.query_command(self.target_db_alias,
                                                                               echo_queries=True)):
                    return False
            else:
                upsery_query = f"""
INSERT INTO {self.target_table}
SELECT {self.target_table}_upsert.*
FROM {self.target_table}_upsert
ON CONFLICT ({key_definition})
DO UPDATE SET {set_clause}"""
                if not shell.run_shell_command(f'echo {shlex.quote(upsery_query)} \\\n  | '
                                               + mara_db.shell.query_command(self.target_db_alias, echo_queries=True)):
                    return False

        # update data_integration_incremental_copy_status
        incremental_copy_status.update(self.node_path(), self.source_db_alias,
                                       self.source_table, max_modification_value)
        return True

    def _copy_command(self, target_table, replace):
        """Helper function for creating the actual copy command"""
        return (_SQLCommand.shell_command(self)
                + '  | ' + shell.sed_command(replace)
                + '  | ' + mara_db.shell.copy_command(self.source_db_alias, self.target_db_alias,
                                                      target_table, timezone=self.timezone,
                                                      csv_format=self.csv_format, delimiter_char=self.delimiter_char))

    def html_doc_items(self) -> [(str, str)]:
        return [('source db', _.tt[self.source_db_alias]),
                ('source table', _.tt[self.source_table]),
                ('modification comparison', _.tt[self.modification_comparison])] \
               + _SQLCommand.html_doc_items(self, self.source_db_alias) \
               + [('comparison value placeholder', _.tt[self.comparison_value_placeholder]),
                  ('target db', _.tt[self.target_db_alias]),
                  ('target table', _.tt[self.target_table]),
                  ('primary_keys', _.tt[repr(self.primary_keys)]),
                  ('timezone', _.tt[self.timezone or '']),
                  ('csv format', _.tt[self.csv_format or '']),
                  ('delimiter char', _.tt[self.delimiter_char or '']),
                  ('use explicit upsert', _.tt[repr(self.use_explicit_upsert)])]


def _expand_pattern_substitution(replace: {str: str}) -> {str: str}:
    """Helper function for replacing callables with their value in a dictionary"""
    return {k: (str(v()) if callable(v) else str(v)) for k, v in replace.items()}


@functools.singledispatch
def _sql_syntax_higlighting_lexter(db):
    """Returns the best lexer from http://pygments.org/docs/lexers/ for a database dialect"""
    return 'sql'


_sql_syntax_higlighting_lexter.register(str, lambda alias: _sql_syntax_higlighting_lexter(mara_db.dbs.db(alias)))
_sql_syntax_higlighting_lexter.register(mara_db.dbs.PostgreSQLDB, lambda _: 'postgresql')
_sql_syntax_higlighting_lexter.register(mara_db.dbs.MysqlDB, lambda _: 'mysql')



from setuptools import setup, find_packages
import re

def get_long_description():
    with open('README.md') as f:
        return re.sub('!\[(.*?)\]\(docs/(.*?)\)', r'![\1](https://github.com/mara/data-integration/raw/master/docs/\2)', f.read())

setup(
    name='data-integration',
    version='2.7.0',

    description='Opinionated lightweight ETL pipeline framework',

    long_description=get_long_description(),
    long_description_content_type='text/markdown',

    url = 'https://github.com/mara/data-integration',

    install_requires=[
        'mara-db>=4.2.0',
        'mara-page>=1.3.0',
        'graphviz>=0.8',
        'python-dateutil>=2.6.1',
        'pythondialog>=3.4.0',
        'more-itertools>=3.1.0',
        'psutil>=5.4.0',
        'wheel>=0.31',
        'requests>=2.19.1'
    ],

    python_requires='>=3.6',

    packages=find_packages(),

    author='Mara contributors',
    license='MIT',

    entry_points={},
)




"""Commands for reading files"""

import json
import pathlib
import shlex
import sys

import enum

import mara_db.dbs
import mara_db.shell
from data_integration.commands import sql
from mara_page import _, html
from .. import config, pipelines


class Compression(enum.EnumMeta):
    """Different compression formats that are understood by file readers"""
    NONE = 'none'
    GZIP = 'gzip'
    TAR_GZIP = 'tar.gzip'
    ZIP = 'zip'


def uncompressor(compression: Compression) -> str:
    """Maps compression methods to command line programs that can unpack the respective files"""
    return {Compression.NONE: 'cat',
            Compression.ZIP: 'unzip -p',
            Compression.GZIP: 'gunzip -d -c',
            Compression.TAR_GZIP: 'tar -xOzf'}[compression]


class ReadFile(pipelines.Command):
    """Reads data from a local file"""

    def __init__(self, file_name: str, compression: Compression, target_table: str,
                 mapper_script_file_name: str = None, make_unique: bool = False,
                 db_alias: str = None, csv_format: bool = False, skip_header: bool = False,
                 delimiter_char: str = None, quote_char: str = None,
                 null_value_string: str = None, timezone: str = None) -> None:
        super().__init__()
        self.file_name = file_name
        self.compression = compression
        self.mapper_script_file_name = mapper_script_file_name
        self.make_unique = make_unique

        self.target_table = target_table
        self.csv_format = csv_format
        self.skip_header = skip_header
        self._db_alias = db_alias
        self.delimiter_char = delimiter_char
        self.quote_char = quote_char
        self.null_value_string = null_value_string
        self.timezone = timezone

    def db_alias(self):
        return self._db_alias or config.default_db_alias()

    def shell_command(self):
        return \
            f'{uncompressor(self.compression)} "{pathlib.Path(config.data_dir()) / self.file_name}" \\\n' \
            + (f'  | {shlex.quote(sys.executable)} "{self.mapper_file_path()}" \\\n'
               if self.mapper_script_file_name else '') \
            + ('  | sort -u \\\n' if self.make_unique else '') \
            + '  | ' + mara_db.shell.copy_from_stdin_command(
                self.db_alias(), csv_format=self.csv_format, target_table=self.target_table,
                skip_header=self.skip_header,
                delimiter_char=self.delimiter_char, quote_char=self.quote_char,
                null_value_string=self.null_value_string, timezone=self.timezone)

    def mapper_file_path(self):
        return self.parent.parent.base_path() / self.mapper_script_file_name

    def html_doc_items(self) -> [(str, str)]:
        return [('file name', _.i[self.file_name]),
                ('compression', _.tt[self.compression]),
                ('mapper script file name', _.i[self.mapper_script_file_name]),
                (_.i['content'], html.highlight_syntax(self.mapper_file_path().read_text().strip('\n')
                                                       if self.mapper_script_file_name and self.mapper_file_path().exists()
                                                       else '', 'python')),
                ('make unique', _.tt[self.make_unique]),
                ('target_table', _.tt[self.target_table]),
                ('db alias', _.tt[self.db_alias()]),
                ('csv format', _.tt[self.csv_format]),
                ('skip header', _.tt[self.skip_header]),
                ('delimiter char',
                 _.tt[json.dumps(self.delimiter_char) if self.delimiter_char != None else None]),
                ('quote char', _.tt[json.dumps(self.quote_char) if self.quote_char != None else None]),
                ('null value string',
                 _.tt[json.dumps(self.null_value_string) if self.null_value_string != None else None]),
                ('time zone', _.tt[self.timezone]),
                (_.i['shell command'], html.highlight_syntax(self.shell_command(), 'bash'))]


class ReadSQLite(sql._SQLCommand):
    def __init__(self, sqlite_file_name: str, target_table: str,
                 sql_statement: str = None, sql_file_name: str = None, replace: {str: str} = None,
                 db_alias: str = None, timezone: str = None) -> None:
        sql._SQLCommand.__init__(self, sql_statement, sql_file_name, replace)
        self.sqlite_file_name = sqlite_file_name

        self.target_table = target_table
        self._db_alias = db_alias
        self.timezone = timezone

    @property
    def db_alias(self):
        return self._db_alias or config.default_db_alias()

    def shell_command(self):
        return (sql._SQLCommand.shell_command(self)
                + '  | ' + mara_db.shell.copy_command(
                    mara_db.dbs.SQLiteDB(file_name=config.data_dir().absolute() / self.sqlite_file_name),
                    self.db_alias, self.target_table, timezone=self.timezone))

    def html_doc_items(self) -> [(str, str)]:
        return [('sqlite file name', _.i[self.sqlite_file_name])] \
               + sql._SQLCommand.html_doc_items(self, None) \
               + [('target_table', _.tt[self.target_table]),
                  ('db alias', _.tt[self.db_alias]),
                  ('time zone', _.tt[self.timezone]),
                  (_.i['shell command'], html.highlight_syntax(self.shell_command(), 'bash'))]


class ReadScriptOutput(pipelines.Command):
    """Reads the output from a python script into a database table"""

    def __init__(self, file_name: str, target_table: str, make_unique: bool = False,
                 db_alias: str = None, csv_format: bool = False, skip_header: bool = False,
                 delimiter_char: str = None, quote_char: str = None,
                 null_value_string: str = None, timezone: str = None) -> None:
        super().__init__()
        self.file_name = file_name
        self.make_unique = make_unique

        self.target_table = target_table
        self.csv_format = csv_format
        self.skip_header = skip_header
        self._db_alias = db_alias
        self.delimiter_char = delimiter_char
        self.quote_char = quote_char
        self.null_value_string = null_value_string
        self.timezone = timezone

    def db_alias(self):
        return self._db_alias or config.default_db_alias()

    def shell_command(self):
        return f'{shlex.quote(sys.executable)} "{self.file_path()}" \\\n' \
               + ('  | sort -u \\\n' if self.make_unique else '') \
               + '  | ' + mara_db.shell.copy_from_stdin_command(
            self.db_alias(), csv_format=self.csv_format, target_table=self.target_table, skip_header=self.skip_header,
            delimiter_char=self.delimiter_char, quote_char=self.quote_char,
            null_value_string=self.null_value_string, timezone=self.timezone)

    def file_path(self):
        return self.parent.parent.base_path() / self.file_name

    def html_doc_items(self) -> [(str, str)]:
        return [('file name', _.i[self.file_name]),
                (_.i['content'], html.highlight_syntax(self.file_path().read_text().strip('\n')
                                                       if self.file_name and self.file_path().exists()
                                                       else '', 'python')),
                ('make unique', _.tt[self.make_unique]),
                ('target_table', _.tt[self.target_table]),
                ('db alias', _.tt[self.db_alias()]),
                ('delimiter char',
                 _.tt[json.dumps(self.delimiter_char) if self.delimiter_char != None else None]),
                ('quote char', _.tt[json.dumps(self.quote_char) if self.quote_char != None else None]),
                ('null value string',
                 _.tt[json.dumps(self.null_value_string) if self.null_value_string != None else None]),
                ('time zone', _.tt[self.timezone]),
                (_.i['shell command'], html.highlight_syntax(self.shell_command(), 'bash'))]



"""Command execution in bash shells"""

import time

from . import config
from .logging import logger


def run_shell_command(command: str, log_command: bool = True):
    """
    Runs a command in a bash shell and logs the output of the command in (near)real-time.

    Args:
        command: The command to run
        log_command: When true, then the command itself is logged before execution

    Returns:
        Either (in order)
        - False when the exit code of the command was not 0
        - True when there was no output to stdout
        - The output to stdout, as an array of lines
    """
    import shlex, subprocess, threading

    if log_command:
        logger.log(command, format=logger.Format.ITALICS)

    process = subprocess.Popen(shlex.split(config.bash_command_string()) + ['-c', command],
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               universal_newlines=True)

    # keep stdout output
    output_lines = []

    # unfortunately, only file descriptors and the system stream can be passed to
    # subprocess.Popen(..) (and not custom streams without a file handle).
    # So in order to see be able to log the output in real-time, we have to
    # query the output steams of the process from to separate threads
    def read_process_output():
        for line in process.stdout:
            output_lines.append(line)
            logger.log(line, format=logger.Format.VERBATIM)

        for line in process.stderr:
            logger.log(line, format=logger.Format.VERBATIM, is_error=True)

    read_output_thread = threading.Thread(target=read_process_output)
    read_output_thread.start()

    # wait until the process finishes
    while process.poll() is None:
        time.sleep(0.005)

    read_output_thread.join()

    exitcode = process.returncode
    if exitcode != 0:
        logger.log(f'exit code {exitcode}', is_error=True, format=logger.Format.ITALICS)
        return False

    return output_lines or True


def sed_command(replace: {str: str}) -> str:
    """
    Creates a sed command string from a dictionary of replacements

    Examples:
    >>> print(sed_command({'foo':'a','bar':'b'}))
    sed "s/foo/a/g; s/bar/b/g"
    """

    def quote(s):
        return str(s).replace('/', '\/').replace('"', '\\\"').replace('\n', '\\\\\n')

    return 'sed "' + \
           ';'.join(
               ['s/' + quote(search) + '/' + quote(_replace) + '/g' for search, _replace in
                replace.items()]) \
           + '"'


if __name__ == "__main__":
    run_shell_command('ping -c 3 google.com; ping null')
